@@ -218,6 +218,7 @@ static const char *opts_file = ACM_CONF_DIR "/" ACM_OPTS_FILE;
 static const char *addr_file = ACM_CONF_DIR "/" ACM_ADDR_FILE;
 static char log_file[128] = IBACM_LOG_FILE;
 static int log_level = 0;
+static int umad_debug_level;
 static char lock_file[128] = IBACM_PID_FILE;
 static short server_port = 6125;
 static int server_mode = IBACM_SERVER_MODE_DEFAULT;
@@ -2082,11 +2083,19 @@ __acm_ep_insert_addr(struct acmc_ep *ep, const char *name, uint8_t *addr,
 		;
 	if (i == ep->nmbr_ep_addrs) {
 		struct acmc_addr *new_info;
+		int j;
+
 		new_info = realloc(ep->addr_info, (i + 1) * sizeof(*ep->addr_info));
 		if (!new_info) {
 			ret = ENOMEM;
 			goto out;
 		}
+
+		/* id_string needs to point to the reallocated string_buf */
+		for (j = 0; (j < ep->nmbr_ep_addrs); j++) {
+			new_info[j].addr.id_string = new_info[j].string_buf;
+		}
+
 		ep->addr_info = new_info;
 
 		/* Added memory is not initialized */
@@ -2108,7 +2117,9 @@ __acm_ep_insert_addr(struct acmc_ep *ep, const char *name, uint8_t *addr,
 		}
 	}
 	ep->addr_info[i].addr.type = addr_type;
-	strncpy(ep->addr_info[i].string_buf, name, ACM_MAX_ADDRESS);
+	if (!check_snprintf(ep->addr_info[i].string_buf,
+			    sizeof(ep->addr_info[i].string_buf), "%s", name))
+		return EINVAL;
 	memcpy(ep->addr_info[i].addr.info.addr, tmp, ACM_MAX_ADDRESS);
 	ret = ep->port->prov->add_address(&ep->addr_info[i].addr,
 					  ep->prov_ep_context,
@@ -2824,6 +2835,17 @@ static void acm_load_prov_config(void)
 	}
 }
 
+static int acm_string_end_compare(const char *s1, const char *s2)
+{
+	size_t s1_len = strlen(s1);
+	size_t s2_len = strlen(s2);
+
+	if (s1_len < s2_len)
+		return -1;
+
+	return strcmp(s1 + s1_len - s2_len, s2);
+}
+
 static int acm_open_providers(void)
 {
 	DIR *shlib_dir;
@@ -2846,7 +2868,7 @@ static int acm_open_providers(void)
 	}
 
 	while ((dent = readdir(shlib_dir))) {
-		if (!strstr(dent->d_name, ".so"))
+		if (acm_string_end_compare(dent->d_name, ".so"))
 			continue;
 
 		if (!check_snprintf(file_name, sizeof(file_name), "%s/%s",
@@ -3069,9 +3091,10 @@ static void acmc_recv_mad(struct acmc_port *port)
 	}
 
 	hdr = &resp.sa_mad.mad_hdr;
-	acm_log(2, "bv %x cls %x cv %x mtd %x st %d tid %" PRIx64 "x at %x atm %x\n",
+	acm_log(2, "bv %x cls %x cv %x mtd %x st %d tid %" PRIx64 " at %x atm %x\n",
 		hdr->base_version, hdr->mgmt_class, hdr->class_version,
-		hdr->method, hdr->status, be64toh(hdr->tid), hdr->attr_id, hdr->attr_mod);
+		hdr->method, be16toh(hdr->status), be64toh(hdr->tid),
+		be16toh(hdr->attr_id), be32toh(hdr->attr_mod));
 	found = 0;
 	pthread_mutex_lock(&port->lock);
 	list_for_each(&port->sa_pending, req, entry) {
@@ -3167,6 +3190,11 @@ static void acm_set_options(void)
 			strcpy(log_file, value);
 		else if (!strcasecmp("log_level", opt))
 			log_level = atoi(value);
+		else if (!strcasecmp("umad_debug_level", opt)) {
+			umad_debug_level = atoi(value);
+			if (umad_debug_level > 0)
+				umad_debug(umad_debug_level);
+		}
 		else if (!strcasecmp("lock_file", opt))
 			strcpy(lock_file, value);
 		else if (!strcasecmp("server_port", opt))
@@ -3208,6 +3236,7 @@ static void acm_log_options(void)
 
 	acm_log(0, "log file %s\n", log_file);
 	acm_log(0, "log level %d\n", log_level);
+	acm_log(0, "umad debug level %d\n", umad_debug_level);
 	acm_log(0, "lock file %s\n", lock_file);
 	acm_log(0, "server_port %d\n", server_port);
 	acm_log(0, "server_mode %s\n", server_mode_names[server_mode]);
