@@ -103,12 +103,12 @@ static unsigned int get_abi_version(void)
 	    0) {
 		IBWARN("can't read ABI version from %s/%s (%m): is ib_umad module loaded?",
 		       IB_UMAD_ABI_DIR, IB_UMAD_ABI_FILE);
-		abi_version = 1 << 31;
+		abi_version = (1U) << 31;
 		return 0;
 	}
 
 	if (abi_version < IB_UMAD_ABI_VERSION) {
-		abi_version = 1 << 31;
+		abi_version = (1U) << 31;
 		return 0;
 	}
 
@@ -148,7 +148,7 @@ static int get_port(const char *ca_name, const char *dir, int portnum, umad_port
 {
 	char port_dir[256];
 	union umad_gid gid;
-	struct dirent **namelist = NULL;
+	struct dirent **namelist;
 	int i, len, num_pkeys = 0;
 	uint32_t capmask;
 
@@ -158,23 +158,24 @@ static int get_port(const char *ca_name, const char *dir, int portnum, umad_port
 
 	len = snprintf(port_dir, sizeof(port_dir), "%s/%d", dir, portnum);
 	if (len < 0 || len > sizeof(port_dir))
-		goto clean;
+		return -EIO;
 
 	if (sys_read_uint(port_dir, SYS_PORT_LMC, &port->lmc) < 0)
-		goto clean;
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_SMLID, &port->sm_lid) < 0)
-		goto clean;
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_SMSL, &port->sm_sl) < 0)
-		goto clean;
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_LID, &port->base_lid) < 0)
-		goto clean;
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_STATE, &port->state) < 0)
-		goto clean;
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_PHY_STATE, &port->phys_state) < 0)
-		goto clean;
-	sys_read_uint(port_dir, SYS_PORT_RATE, &port->rate);
+		return -EIO;
+	if (sys_read_uint(port_dir, SYS_PORT_RATE, &port->rate) < 0)
+		return -EIO;
 	if (sys_read_uint(port_dir, SYS_PORT_CAPMASK, &capmask) < 0)
-		goto clean;
+		return -EIO;
 
 	if (sys_read_string(port_dir, SYS_PORT_LINK_LAYER,
 	    port->link_layer, UMAD_CA_NAME_LEN) < 0)
@@ -184,7 +185,7 @@ static int get_port(const char *ca_name, const char *dir, int portnum, umad_port
 	port->capmask = htobe32(capmask);
 
 	if (sys_read_gid(port_dir, SYS_PORT_GID, &gid) < 0)
-		goto clean;
+		return -EIO;
 
 	port->gid_prefix = gid.global.subnet_prefix;
 	port->port_guid = gid.global.interface_id;
@@ -194,37 +195,37 @@ static int get_port(const char *ca_name, const char *dir, int portnum, umad_port
 	if (num_pkeys <= 0) {
 		IBWARN("no pkeys found for %s:%u (at dir %s)...",
 		       port->ca_name, port->portnum, port_dir);
-		goto clean;
+		return -EIO;
 	}
 	port->pkeys = calloc(num_pkeys, sizeof(port->pkeys[0]));
 	if (!port->pkeys) {
 		IBWARN("get_port: calloc failed: %s", strerror(errno));
-		goto clean;
+		goto clean_names;
 	}
 	for (i = 0; i < num_pkeys; i++) {
 		unsigned idx, val;
+
 		idx = strtoul(namelist[i]->d_name, NULL, 0);
-		sys_read_uint(port_dir, namelist[i]->d_name, &val);
+		if (sys_read_uint(port_dir, namelist[i]->d_name, &val) < 0)
+			goto clean_pkeys;
 		port->pkeys[idx] = val;
-		free(namelist[i]);
 	}
 	port->pkeys_size = num_pkeys;
+	for (i = 0; i < num_pkeys; i++)
+		free(namelist[i]);
 	free(namelist);
-	namelist = NULL;
 	port_dir[len] = '\0';
 
 	/* FIXME: handle gids */
 
 	return 0;
 
-clean:
-	if (namelist) {
-		for (i = 0; i < num_pkeys; i++)
-			free(namelist[i]);
-		free(namelist);
-	}
-	if (port->pkeys)
-		free(port->pkeys);
+clean_pkeys:
+	free(port->pkeys);
+clean_names:
+	for (i = 0; i < num_pkeys; i++)
+		free(namelist[i]);
+	free(namelist);
 	return -EIO;
 }
 
@@ -551,6 +552,13 @@ static int dev_to_umad_id(const char *dev, unsigned port)
 	return -1;		/* not found */
 }
 
+static int umad_ca_device_list_compare_function(const void *node_a,
+						const void *node_b)
+{
+	return strcmp((*((const struct umad_device_node **)node_a))->ca_name,
+		      (*((const struct umad_device_node **)node_b))->ca_name);
+}
+
 /*******************************
  * Public interface
  */
@@ -1207,6 +1215,53 @@ void umad_dump(void *umad)
 	umad_addr_dump(&mad->addr);
 }
 
+int umad_sort_ca_device_list(struct umad_device_node **head, size_t size)
+{
+	int errsv = 0;
+	size_t i;
+	struct umad_device_node *node;
+	struct umad_device_node **nodes_array = NULL;
+
+	if (!size)
+		for (node = *head; node; node = node->next)
+			size++;
+	if (size < 2)
+		return 0;
+
+	nodes_array = calloc(size, sizeof(struct umad_device_node *));
+	if (!nodes_array) {
+		errsv = ENOMEM;
+		goto exit;
+	}
+
+	node = *head;
+	for (i = 0; i < size; i++) {
+		if (!node) {
+			errsv = EINVAL;
+			goto exit;
+		}
+		nodes_array[i] = node;
+		node = node->next;
+	}
+	if (node) {
+		errsv = EINVAL;
+		goto exit;
+	}
+
+	qsort(nodes_array, size, sizeof(struct umad_device_node *),
+	      umad_ca_device_list_compare_function);
+
+	for (i = 0; i < size - 1; i++)
+		nodes_array[i]->next = nodes_array[i + 1];
+
+	*head = nodes_array[0];
+	nodes_array[size - 1]->next = NULL;
+exit:
+	free(nodes_array);
+
+	return errsv;
+}
+
 struct umad_device_node *umad_get_ca_device_list(void)
 {
 	DIR *dir;
