@@ -198,7 +198,7 @@ static void print_formated_gid(union ibv_gid *gid, int i,
 
 static int print_all_port_gids(struct ibv_context *ctx,
 			       struct ibv_port_attr *port_attr,
-			       uint8_t port_num)
+			       uint32_t port_num)
 {
 	enum ibv_gid_type_sysfs type;
 	union ibv_gid gid;
@@ -210,7 +210,7 @@ static int print_all_port_gids(struct ibv_context *ctx,
 	for (i = 0; i < tbl_len; i++) {
 		rc = ibv_query_gid(ctx, port_num, i, &gid);
 		if (rc) {
-			fprintf(stderr, "Failed to query gid to port %d, index %d\n",
+			fprintf(stderr, "Failed to query gid to port %u, index %d\n",
 			       port_num, i);
 			return rc;
 		}
@@ -499,7 +499,7 @@ static int print_hca_cap(struct ibv_device *ib_dev, uint8_t ib_port)
 	struct ibv_device_attr_ex device_attr = {};
 	struct ibv_port_attr port_attr;
 	int rc = 0;
-	uint8_t port;
+	uint32_t port;
 	char buf[256];
 
 	ctx = ibv_open_device(ib_dev);
@@ -618,7 +618,7 @@ static int print_hca_cap(struct ibv_device *ib_dev, uint8_t ib_port)
 			fprintf(stderr, "Failed to query port %u props\n", port);
 			goto cleanup;
 		}
-		printf("\t\tport:\t%d\n", port);
+		printf("\t\tport:\t%u\n", port);
 		printf("\t\t\tstate:\t\t\t%s (%d)\n",
 		       port_state_str(port_attr.state), port_attr.state);
 		printf("\t\t\tmax_mtu:\t\t%s (%d)\n",
