@@ -1,6 +1,6 @@
 /* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
 /*
- * Copyright 2019 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright 2019-2020 Amazon.com, Inc. or its affiliates. All rights reserved.
  */
 
 #ifndef __EFA_VERBS_H__
@@ -9,7 +9,7 @@
 #include <infiniband/driver.h>
 #include <infiniband/verbs.h>
 
-int efa_query_device(struct ibv_context *uctx, struct ibv_device_attr *attr);
+int efa_query_device_ctx(struct efa_context *ctx);
 int efa_query_port(struct ibv_context *uctx, uint8_t port,
 		   struct ibv_port_attr *attr);
 int efa_query_device_ex(struct ibv_context *context,
@@ -23,11 +23,15 @@ int efa_dereg_mr(struct verbs_mr *vmr);
 
 struct ibv_cq *efa_create_cq(struct ibv_context *uctx, int ncqe,
 			     struct ibv_comp_channel *ch, int vec);
+struct ibv_cq_ex *efa_create_cq_ex(struct ibv_context *uctx,
+				   struct ibv_cq_init_attr_ex *attr_ex);
 int efa_destroy_cq(struct ibv_cq *ibvcq);
 int efa_poll_cq(struct ibv_cq *ibvcq, int nwc, struct ibv_wc *wc);
 
 struct ibv_qp *efa_create_qp(struct ibv_pd *ibvpd,
 			     struct ibv_qp_init_attr *attr);
+struct ibv_qp *efa_create_qp_ex(struct ibv_context *ibvctx,
+				struct ibv_qp_init_attr_ex *attr_ex);
 int efa_modify_qp(struct ibv_qp *ibvqp, struct ibv_qp_attr *attr,
 		  int ibv_qp_attr_mask);
 int efa_query_qp(struct ibv_qp *ibvqp, struct ibv_qp_attr *attr, int attr_mask,
