@@ -589,10 +589,8 @@ static int hns_roce_alloc_recv_inl_buf(struct ibv_qp_cap *cap,
 
 	qp->rq_rinl_buf.wqe_list[0].sg_list = calloc(cnt * cap->max_recv_sge,
 					sizeof(struct hns_roce_rinl_sge));
-	if (!qp->rq_rinl_buf.wqe_list[0].sg_list) {
-		free(qp->rq_rinl_buf.wqe_list);
+	if (!qp->rq_rinl_buf.wqe_list[0].sg_list)
 		return ENOMEM;
-	}
 
 	for (i = 0; i < cnt; i++) {
 		int wqe_size = i * cap->max_recv_sge;
@@ -713,7 +711,12 @@ static void set_extend_sge_param(struct hns_roce_device *hr_dev,
 	}
 
 	qp->ex_sge.sge_shift = HNS_ROCE_SGE_SHIFT;
-	qp->ex_sge.sge_cnt = cnt;
+
+	/* If the number of extended sge is not zero, they MUST use the
+	 * space of HNS_HW_PAGE_SIZE at least.
+	 */
+	qp->ex_sge.sge_cnt = cnt ?
+			     max(cnt, HNS_HW_PAGE_SIZE / HNS_ROCE_SGE_SIZE) : 0;
 }
 
 static void hns_roce_set_qp_params(struct ibv_pd *pd,
