@@ -127,15 +127,17 @@ static bool dr_mask_is_tnl_gre_set(struct dr_match_misc *misc)
 		misc->gre_k_present || misc->gre_s_present);
 }
 
-#define DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, gre_udp) ( \
-	(_misc2).outer_first_mpls_over_##gre_udp##_label || \
-	(_misc2).outer_first_mpls_over_##gre_udp##_exp || \
-	(_misc2).outer_first_mpls_over_##gre_udp##_s_bos || \
-	(_misc2).outer_first_mpls_over_##gre_udp##_ttl)
+#define DR_MASK_IS_OUTER_MPLS_OVER_GRE_SET(_misc) (\
+	(_misc)->outer_first_mpls_over_gre_label || \
+	(_misc)->outer_first_mpls_over_gre_exp || \
+	(_misc)->outer_first_mpls_over_gre_s_bos || \
+	(_misc)->outer_first_mpls_over_gre_ttl)
 
-#define DR_MASK_IS_TNL_MPLS_SET(_misc2) ( \
-	DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, gre) || \
-	DR_MASK_IS_OUTER_MPLS_OVER_GRE_UDP_SET(_misc2, udp))
+#define DR_MASK_IS_OUTER_MPLS_OVER_UDP_SET(_misc) (\
+	(_misc)->outer_first_mpls_over_udp_label || \
+	(_misc)->outer_first_mpls_over_udp_exp || \
+	(_misc)->outer_first_mpls_over_udp_s_bos || \
+	(_misc)->outer_first_mpls_over_udp_ttl)
 
 static bool
 dr_mask_is_vxlan_gpe_set(struct dr_match_misc3 *misc3)
@@ -148,8 +150,8 @@ dr_mask_is_vxlan_gpe_set(struct dr_match_misc3 *misc3)
 static bool
 dr_matcher_supp_vxlan_gpe(struct dr_devx_caps *caps)
 {
-	return caps->flex_protocols &
-	       MLX5_FLEX_PARSER_VXLAN_GPE_ENABLED;
+	return (caps->sw_format_ver == MLX5_HW_CONNECTX_6DX) ||
+	       (caps->flex_protocols & MLX5_FLEX_PARSER_VXLAN_GPE_ENABLED);
 }
 
 static bool
@@ -168,11 +170,23 @@ static bool dr_mask_is_tnl_geneve_set(struct dr_match_misc *misc)
 	       misc->geneve_opt_len;
 }
 
+static int dr_matcher_supp_geneve_tlv_option(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_GENEVE_OPT_0_ENABLED;
+}
+
+static bool dr_mask_is_tnl_geneve_tlv_opt(struct dr_match_param *mask,
+					  struct mlx5dv_dr_domain *dmn)
+{
+	return mask->misc3.geneve_tlv_option_0_data &&
+	       dr_matcher_supp_geneve_tlv_option(&dmn->info.caps);
+}
+
 static bool
 dr_matcher_supp_tnl_geneve(struct dr_devx_caps *caps)
 {
-	return caps->flex_protocols &
-	       MLX5_FLEX_PARSER_GENEVE_ENABLED;
+	return (caps->sw_format_ver == MLX5_HW_CONNECTX_6DX) ||
+	       (caps->flex_protocols & MLX5_FLEX_PARSER_GENEVE_ENABLED);
 }
 
 static bool
@@ -202,12 +216,14 @@ static bool dr_mask_is_tnl_gtpu(struct dr_match_param *mask,
 
 static inline int dr_matcher_supp_icmp_v4(struct dr_devx_caps *caps)
 {
-	return caps->flex_protocols & MLX5_FLEX_PARSER_ICMP_V4_ENABLED;
+	return (caps->sw_format_ver == MLX5_HW_CONNECTX_6DX) ||
+	       (caps->flex_protocols & MLX5_FLEX_PARSER_ICMP_V4_ENABLED);
 }
 
 static inline int dr_matcher_supp_icmp_v6(struct dr_devx_caps *caps)
 {
-	return caps->flex_protocols & MLX5_FLEX_PARSER_ICMP_V6_ENABLED;
+	return (caps->sw_format_ver == MLX5_HW_CONNECTX_6DX) ||
+	       (caps->flex_protocols & MLX5_FLEX_PARSER_ICMP_V6_ENABLED);
 }
 
 static bool dr_mask_is_icmpv6_set(struct dr_match_misc3 *misc3)
@@ -249,12 +265,73 @@ static bool dr_mask_is_gvmi_or_qpn_set(struct dr_match_misc *misc)
 	return (misc->source_sqn || misc->source_port);
 }
 
+static bool dr_mask_is_flex_parser_id_0_3_set(uint32_t flex_parser_id,
+					      uint32_t flex_parser_value)
+{
+	if (flex_parser_id)
+		return flex_parser_id <= DR_STE_MAX_FLEX_0_ID;
+
+	/* Using flex_parser 0 means that id is zero, thus value must be set. */
+	return flex_parser_value;
+}
+
+static bool dr_mask_is_flex_parser_0_3_set(struct dr_match_misc4 *misc4)
+{
+	return (dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_0,
+			misc4->prog_sample_field_value_0) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_1,
+			misc4->prog_sample_field_value_1) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_2,
+			misc4->prog_sample_field_value_2) ||
+		dr_mask_is_flex_parser_id_0_3_set(misc4->prog_sample_field_id_3,
+			misc4->prog_sample_field_value_3));
+}
+
+static bool dr_mask_is_flex_parser_id_4_7_set(uint32_t flex_parser_id)
+{
+	return flex_parser_id > DR_STE_MAX_FLEX_0_ID &&
+	       flex_parser_id <= DR_STE_MAX_FLEX_1_ID;
+}
+
+static bool dr_mask_is_flex_parser_4_7_set(struct dr_match_misc4 *misc4)
+{
+	return (dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_0) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_1) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_2) ||
+		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_3));
+}
+
+static int dr_matcher_supp_tnl_mpls_over_gre(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_MPLS_OVER_GRE_ENABLED;
+}
+
+static bool dr_mask_is_tnl_mpls_over_gre(struct dr_match_param *mask,
+					 struct mlx5dv_dr_domain *dmn)
+{
+	return DR_MASK_IS_OUTER_MPLS_OVER_GRE_SET(&mask->misc2) &&
+	       dr_matcher_supp_tnl_mpls_over_gre(&dmn->info.caps);
+}
+
+static int dr_matcher_supp_tnl_mpls_over_udp(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & mlx5_FLEX_PARSER_MPLS_OVER_UDP_ENABLED;
+}
+
+static bool dr_mask_is_tnl_mpls_over_udp(struct dr_match_param *mask,
+					 struct mlx5dv_dr_domain *dmn)
+{
+	return DR_MASK_IS_OUTER_MPLS_OVER_UDP_SET(&mask->misc2) &&
+	       dr_matcher_supp_tnl_mpls_over_udp(&dmn->info.caps);
+}
+
 static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 				       struct dr_matcher_rx_tx *nic_matcher)
 {
 	struct dr_domain_rx_tx *nic_dmn = nic_matcher->nic_tbl->nic_dmn;
 	struct dr_ste_build *sb = nic_matcher->ste_builder;
 	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	struct dr_ste_ctx *ste_ctx = dmn->ste_ctx;
 	struct dr_match_param mask = {};
 	bool allow_empty_match = false;
 	bool inner, rx;
@@ -280,6 +357,9 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC3)
 		mask.misc3 = matcher->mask.misc3;
 
+	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC4)
+		mask.misc4 = matcher->mask.misc4;
+
 	ret = dr_ste_build_pre_check(dmn, matcher->match_criteria,
 				     &matcher->mask, NULL);
 	if (ret)
@@ -303,81 +383,101 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 		ipv = mask.outer.ip_version;
 
 		if (dr_mask_is_wqe_metadata_set(&mask.misc2))
-			dr_ste_build_general_purpose(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_general_purpose(ste_ctx, &sb[idx++],
+						     &mask, inner, rx);
 
 		if (dr_mask_is_reg_c_0_3_set(&mask.misc2))
-			dr_ste_build_register_0(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_register_0(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (dr_mask_is_reg_c_4_7_set(&mask.misc2))
-			dr_ste_build_register_1(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_register_1(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (dr_mask_is_gvmi_or_qpn_set(&mask.misc) &&
 		    (dmn->type == MLX5DV_DR_DOMAIN_TYPE_FDB ||
 		     dmn->type == MLX5DV_DR_DOMAIN_TYPE_NIC_RX))
-			dr_ste_build_src_gvmi_qpn(&sb[idx++], &mask,
-						  &dmn->info.caps,
+			dr_ste_build_src_gvmi_qpn(ste_ctx, &sb[idx++],
+						  &mask, &dmn->info.caps,
 						  inner, rx);
 
 		if (dr_mask_is_smac_set(&mask.outer) &&
 		    dr_mask_is_dmac_set(&mask.outer))
-			dr_ste_build_eth_l2_src_dst(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_src_dst(ste_ctx, &sb[idx++],
+						    &mask, inner, rx);
 
 		if (dr_mask_is_smac_set(&mask.outer))
-			dr_ste_build_eth_l2_src(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_src(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (DR_MASK_IS_L2_DST(mask.outer, mask.misc, outer))
-			dr_ste_build_eth_l2_dst(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_dst(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (ipv == 4) {
 			if (dr_mask_is_ttl_set(&mask.outer))
-				dr_ste_build_eth_l3_ipv4_misc(&sb[idx++], &mask,
-							      inner, rx);
+				dr_ste_build_eth_l3_ipv4_misc(ste_ctx, &sb[idx++],
+							      &mask, inner, rx);
 
 			if (dr_mask_is_ipv4_5_tuple_set(&mask.outer))
-				dr_ste_build_eth_l3_ipv4_5_tuple(&sb[idx++], &mask,
-								 inner, rx);
+				dr_ste_build_eth_l3_ipv4_5_tuple(ste_ctx, &sb[idx++],
+								 &mask, inner, rx);
 		} else if (ipv == 6) {
 			if (dr_mask_is_dst_addr_set(&mask.outer))
-				dr_ste_build_eth_l3_ipv6_dst(&sb[idx++], &mask,
-							     inner, rx);
+				dr_ste_build_eth_l3_ipv6_dst(ste_ctx, &sb[idx++],
+							     &mask, inner, rx);
 
 			if (dr_mask_is_src_addr_set(&mask.outer))
-				dr_ste_build_eth_l3_ipv6_src(&sb[idx++], &mask,
-							     inner, rx);
+				dr_ste_build_eth_l3_ipv6_src(ste_ctx, &sb[idx++],
+							     &mask, inner, rx);
 
 			if (DR_MASK_IS_ETH_L4_SET(mask.outer, mask.misc, outer))
-				dr_ste_build_eth_ipv6_l3_l4(&sb[idx++], &mask,
-							    inner, rx);
+				dr_ste_build_eth_ipv6_l3_l4(ste_ctx, &sb[idx++],
+							    &mask, inner, rx);
 		}
 
-		if (dr_mask_is_tnl_vxlan_gpe(&mask, dmn))
-			dr_ste_build_tnl_vxlan_gpe(&sb[idx++], &mask,
-						   inner, rx);
-		else if (dr_mask_is_tnl_geneve(&mask, dmn))
-			dr_ste_build_tnl_geneve(&sb[idx++], &mask,
-						inner, rx);
-		else if (dr_mask_is_tnl_gtpu(&mask, dmn))
-			dr_ste_build_tnl_gtpu(&sb[idx++], &mask,
-					      inner, rx);
+		if (dr_mask_is_tnl_vxlan_gpe(&mask, dmn)) {
+			dr_ste_build_tnl_vxlan_gpe(ste_ctx, &sb[idx++],
+						   &mask, inner, rx);
+		} else if (dr_mask_is_tnl_geneve(&mask, dmn) ||
+			 dr_mask_is_tnl_geneve_tlv_opt(&mask, dmn)) {
+			if (dr_mask_is_tnl_geneve(&mask, dmn))
+				dr_ste_build_tnl_geneve(ste_ctx, &sb[idx++],
+							&mask, inner, rx);
+
+			if (dr_mask_is_tnl_geneve_tlv_opt(&mask, dmn))
+				dr_ste_build_tnl_geneve_tlv_opt(ste_ctx, &sb[idx++],
+								&mask, &dmn->info.caps,
+								inner, rx);
+		} else if (dr_mask_is_tnl_gtpu(&mask, dmn)) {
+			dr_ste_build_tnl_gtpu(ste_ctx, &sb[idx++],
+					      &mask, inner, rx);
+		}
 
 		if (DR_MASK_IS_ETH_L4_MISC_SET(mask.misc3, outer))
-			dr_ste_build_eth_l4_misc(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l4_misc(ste_ctx, &sb[idx++],
+						 &mask, inner, rx);
 
 		if (DR_MASK_IS_FIRST_MPLS_SET(mask.misc2, outer))
-			dr_ste_build_mpls(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_mpls(ste_ctx, &sb[idx++],
+					  &mask, inner, rx);
 
-		if (DR_MASK_IS_TNL_MPLS_SET(mask.misc2))
-			dr_ste_build_tnl_mpls(&sb[idx++], &mask, inner, rx);
+		if (dr_mask_is_tnl_mpls_over_gre(&mask, dmn))
+			dr_ste_build_tnl_mpls_over_gre(ste_ctx, &sb[idx++],
+						       &mask, &dmn->info.caps,
+						       inner, rx);
+		else if (dr_mask_is_tnl_mpls_over_udp(&mask, dmn))
+			dr_ste_build_tnl_mpls_over_udp(ste_ctx, &sb[idx++],
+						       &mask, &dmn->info.caps,
+						       inner, rx);
+
+		if (dr_mask_is_icmp(&mask, dmn))
+			dr_ste_build_icmp(ste_ctx, &sb[idx++],
+					  &mask, &dmn->info.caps,
+					  inner, rx);
 
-		if (dr_mask_is_icmp(&mask, dmn)) {
-			ret = dr_ste_build_icmp(&sb[idx++],
-						&mask, &dmn->info.caps,
-						inner, rx);
-			if (ret)
-				return ret;
-		}
 		if (dr_mask_is_tnl_gre_set(&mask.misc))
-			dr_ste_build_tnl_gre(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_tnl_gre(ste_ctx, &sb[idx++], &mask, inner, rx);
 	}
 
 	/* Inner */
@@ -389,49 +489,70 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 		ipv = mask.inner.ip_version;
 
 		if (dr_mask_is_eth_l2_tnl_set(&mask.misc))
-			dr_ste_build_eth_l2_tnl(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_tnl(ste_ctx, &sb[idx++], &mask,
+						inner, rx);
 
 		if (dr_mask_is_smac_set(&mask.inner) &&
 		    dr_mask_is_dmac_set(&mask.inner))
-			dr_ste_build_eth_l2_src_dst(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_src_dst(ste_ctx, &sb[idx++],
+						    &mask, inner, rx);
 
 		if (dr_mask_is_smac_set(&mask.inner))
-			dr_ste_build_eth_l2_src(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_src(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (DR_MASK_IS_L2_DST(mask.inner, mask.misc, inner))
-			dr_ste_build_eth_l2_dst(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l2_dst(ste_ctx, &sb[idx++],
+						&mask, inner, rx);
 
 		if (ipv == 4) {
 			if (dr_mask_is_ttl_set(&mask.inner))
-				dr_ste_build_eth_l3_ipv4_misc(&sb[idx++], &mask,
-							      inner, rx);
+				dr_ste_build_eth_l3_ipv4_misc(ste_ctx, &sb[idx++],
+							      &mask, inner, rx);
 
 			if (dr_mask_is_ipv4_5_tuple_set(&mask.inner))
-				dr_ste_build_eth_l3_ipv4_5_tuple(&sb[idx++], &mask,
-								 inner, rx);
+				dr_ste_build_eth_l3_ipv4_5_tuple(ste_ctx, &sb[idx++],
+								 &mask, inner, rx);
 		} else if (ipv == 6) {
 			if (dr_mask_is_dst_addr_set(&mask.inner))
-				dr_ste_build_eth_l3_ipv6_dst(&sb[idx++], &mask,
-							     inner, rx);
+				dr_ste_build_eth_l3_ipv6_dst(ste_ctx, &sb[idx++],
+							     &mask, inner, rx);
 
 			if (dr_mask_is_src_addr_set(&mask.inner))
-				dr_ste_build_eth_l3_ipv6_src(&sb[idx++], &mask,
-							     inner, rx);
+				dr_ste_build_eth_l3_ipv6_src(ste_ctx, &sb[idx++],
+							     &mask,  inner, rx);
 
 			if (DR_MASK_IS_ETH_L4_SET(mask.inner, mask.misc, inner))
-				dr_ste_build_eth_ipv6_l3_l4(&sb[idx++], &mask,
-							    inner, rx);
+				dr_ste_build_eth_ipv6_l3_l4(ste_ctx, &sb[idx++],
+							    &mask, inner, rx);
 		}
 
 		if (DR_MASK_IS_ETH_L4_MISC_SET(mask.misc3, inner))
-			dr_ste_build_eth_l4_misc(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_eth_l4_misc(ste_ctx, &sb[idx++],
+						 &mask, inner, rx);
 
 		if (DR_MASK_IS_FIRST_MPLS_SET(mask.misc2, inner))
-			dr_ste_build_mpls(&sb[idx++], &mask, inner, rx);
+			dr_ste_build_mpls(ste_ctx, &sb[idx++],
+					  &mask, inner, rx);
 
-		if (DR_MASK_IS_TNL_MPLS_SET(mask.misc2))
-			dr_ste_build_tnl_mpls(&sb[idx++], &mask, inner, rx);
+		if (dr_mask_is_tnl_mpls_over_gre(&mask, dmn))
+			dr_ste_build_tnl_mpls_over_gre(ste_ctx, &sb[idx++],
+						       &mask, &dmn->info.caps,
+						       inner, rx);
+		else if (dr_mask_is_tnl_mpls_over_udp(&mask, dmn))
+			dr_ste_build_tnl_mpls_over_udp(ste_ctx, &sb[idx++],
+						       &mask, &dmn->info.caps,
+						       inner, rx);
+	}
 
+	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC4) {
+		if (dr_mask_is_flex_parser_0_3_set(&mask.misc4))
+			dr_ste_build_flex_parser_0(ste_ctx, &sb[idx++],
+						   &mask, false, rx);
+
+		if (dr_mask_is_flex_parser_4_7_set(&mask.misc4))
+			dr_ste_build_flex_parser_1(ste_ctx, &sb[idx++],
+						   &mask, false, rx);
 	}
 
 	/* Empty matcher, takes all */
@@ -772,7 +893,7 @@ mlx5dv_dr_matcher_create(struct mlx5dv_dr_table *tbl,
 	list_node_init(&matcher->matcher_list);
 	list_head_init(&matcher->rule_list);
 
-	pthread_mutex_lock(&tbl->dmn->mutex);
+	dr_domain_lock(tbl->dmn);
 
 	ret = dr_matcher_init(matcher, mask);
 	if (ret)
@@ -782,14 +903,14 @@ mlx5dv_dr_matcher_create(struct mlx5dv_dr_table *tbl,
 	if (ret)
 		goto matcher_uninit;
 
-	pthread_mutex_unlock(&tbl->dmn->mutex);
+	dr_domain_unlock(tbl->dmn);
 
 	return matcher;
 
 matcher_uninit:
 	dr_matcher_uninit(matcher);
 free_matcher:
-	pthread_mutex_unlock(&tbl->dmn->mutex);
+	dr_domain_unlock(tbl->dmn);
 	free(matcher);
 dec_ref:
 	atomic_fetch_sub(&tbl->refcount, 1);
@@ -869,13 +990,14 @@ int mlx5dv_dr_matcher_destroy(struct mlx5dv_dr_matcher *matcher)
 	if (atomic_load(&matcher->refcount) > 1)
 		return EBUSY;
 
-	pthread_mutex_lock(&tbl->dmn->mutex);
+	dr_domain_lock(tbl->dmn);
 
 	dr_matcher_remove_from_tbl(matcher);
 	dr_matcher_uninit(matcher);
 	atomic_fetch_sub(&matcher->tbl->refcount, 1);
 
-	pthread_mutex_unlock(&tbl->dmn->mutex);
+	dr_domain_unlock(tbl->dmn);
+
 	free(matcher);
 
 	return 0;
