@@ -36,7 +36,8 @@
 
 #define DR_RULE_MAX_STE_CHAIN (DR_RULE_MAX_STES + DR_ACTION_MAX_STES)
 
-static int dr_rule_append_to_miss_list(struct dr_ste *new_last_ste,
+static int dr_rule_append_to_miss_list(struct dr_ste_ctx *ste_ctx,
+				       struct dr_ste *new_last_ste,
 				       struct list_head *miss_list,
 				       struct list_head *send_list)
 {
@@ -53,10 +54,13 @@ static int dr_rule_append_to_miss_list(struct dr_ste *new_last_ste,
 		return errno;
 	}
 
-	dr_ste_set_miss_addr(last_ste->hw_ste, dr_ste_get_icm_addr(new_last_ste));
+	dr_ste_set_miss_addr(ste_ctx,
+			     last_ste->hw_ste,
+			     dr_ste_get_icm_addr(new_last_ste));
+
 	list_add_tail(miss_list, &new_last_ste->miss_list_node);
 
-	dr_send_fill_and_append_ste_send_info(last_ste, DR_STE_SIZE_REDUCED,
+	dr_send_fill_and_append_ste_send_info(last_ste, DR_STE_SIZE_CTRL,
 					      0, last_ste->hw_ste,
 					      ste_info_last, send_list, true);
 
@@ -69,6 +73,7 @@ static struct dr_ste
 			       uint8_t *hw_ste)
 {
 	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	struct dr_ste_ctx *ste_ctx = dmn->ste_ctx;
 	struct dr_ste_htbl *new_htbl;
 	struct dr_ste *ste;
 
@@ -84,7 +89,7 @@ static struct dr_ste
 
 	/* One and only entry, never grows */
 	ste = new_htbl->ste_arr;
-	dr_ste_set_miss_addr(hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
+	dr_ste_set_miss_addr(ste_ctx, hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
 	dr_htbl_get(new_htbl);
 
 	return ste;
@@ -104,6 +109,7 @@ static struct dr_ste *dr_rule_create_collision_entry(struct mlx5dv_dr_matcher *m
 	}
 
 	ste->ste_chain_location = orig_ste->ste_chain_location;
+	ste->htbl->pointing_ste = orig_ste->htbl->pointing_ste;
 
 	/* In collision entry, all members share the same miss_list_head */
 	ste->htbl->miss_list = dr_ste_get_miss_list(orig_ste);
@@ -118,7 +124,7 @@ static struct dr_ste *dr_rule_create_collision_entry(struct mlx5dv_dr_matcher *m
 	return ste;
 
 free_tbl:
-	dr_ste_free(ste, matcher, nic_matcher);
+	dr_htbl_put(ste->htbl);
 	return NULL;
 }
 
@@ -128,14 +134,19 @@ static int dr_rule_handle_one_ste_in_update_list(struct dr_ste_send_info *ste_in
 	int ret;
 
 	list_del(&ste_info->send_list);
+
+	/* Copy data to ste, only reduced size or control, the last 16B (mask)
+	 * is already written to the hw.
+	 */
+	if (ste_info->size == DR_STE_SIZE_CTRL)
+		memcpy(ste_info->ste->hw_ste, ste_info->data, DR_STE_SIZE_CTRL);
+	else
+		memcpy(ste_info->ste->hw_ste, ste_info->data, DR_STE_SIZE_REDUCED);
+
 	ret = dr_send_postsend_ste(dmn, ste_info->ste, ste_info->data,
 				   ste_info->size, ste_info->offset);
 	if (ret)
 		goto out;
-	/* Copy data to ste, only reduced size, the last 16B (mask)
-	 * is already written to the hw.
-	 */
-	memcpy(ste_info->ste->hw_ste, ste_info->data, DR_STE_SIZE_REDUCED);
 
 out:
 	free(ste_info);
@@ -190,6 +201,7 @@ dr_rule_rehash_handle_collision(struct mlx5dv_dr_matcher *matcher,
 				struct dr_ste *col_ste,
 				uint8_t *hw_ste)
 {
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
 	struct dr_ste *new_ste;
 	int ret;
 
@@ -197,22 +209,25 @@ dr_rule_rehash_handle_collision(struct mlx5dv_dr_matcher *matcher,
 	if (!new_ste)
 		return NULL;
 
+	/* Update collision pointing STE */
+	new_ste->htbl->pointing_ste = col_ste->htbl->pointing_ste;
+
 	/* In collision entry, all members share the same miss_list_head */
 	new_ste->htbl->miss_list = dr_ste_get_miss_list(col_ste);
 
 	/* Update the previous from the list */
-	ret = dr_rule_append_to_miss_list(new_ste,
+	ret = dr_rule_append_to_miss_list(dmn->ste_ctx, new_ste,
 					  dr_ste_get_miss_list(col_ste),
 					  update_list);
 	if (ret) {
-		dr_dbg(matcher->tbl->dmn, "Failed update dup entry\n");
+		dr_dbg(dmn, "Failed update dup entry\n");
 		goto err_exit;
 	}
 
 	return new_ste;
 
 err_exit:
-	dr_ste_free(new_ste, matcher, nic_matcher);
+	dr_htbl_put(new_ste->htbl);
 	return NULL;
 }
 
@@ -224,7 +239,7 @@ static void dr_rule_rehash_copy_ste_ctrl(struct mlx5dv_dr_matcher *matcher,
 	new_ste->next_htbl = cur_ste->next_htbl;
 	new_ste->ste_chain_location = cur_ste->ste_chain_location;
 
-	if (!dr_ste_is_last_in_rule(nic_matcher, new_ste->ste_chain_location))
+	if (new_ste->next_htbl)
 		new_ste->next_htbl->pointing_ste = new_ste;
 
 	/*
@@ -233,10 +248,8 @@ static void dr_rule_rehash_copy_ste_ctrl(struct mlx5dv_dr_matcher *matcher,
 	 */
 	atomic_init(&new_ste->refcount, atomic_load(&cur_ste->refcount));
 
-	/* Link old STEs rule_mem list to the new ste */
-	dr_rule_update_rule_member(cur_ste, new_ste);
-	list_head_init(&new_ste->rule_list);
-	list_append_list(&new_ste->rule_list, &cur_ste->rule_list);
+	/* Link old STEs rule to the new ste */
+	dr_rule_set_last_member(cur_ste->rule_rx_tx, new_ste, false);
 }
 
 static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
@@ -245,6 +258,7 @@ static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
 					      struct dr_ste_htbl *new_htbl,
 					      struct list_head *update_list)
 {
+	struct dr_ste_ctx *ste_ctx = matcher->tbl->dmn->ste_ctx;
 	uint8_t hw_ste[DR_STE_SIZE] = {};
 	struct dr_ste_send_info *ste_info;
 	bool use_update_list = false;
@@ -258,7 +272,7 @@ static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
 
 	/* Copy STE control and tag */
 	memcpy(hw_ste, cur_ste->hw_ste, DR_STE_SIZE_REDUCED);
-	dr_ste_set_miss_addr(hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
+	dr_ste_set_miss_addr(ste_ctx, hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
 
 	new_idx = dr_ste_calc_hash_index(hw_ste, new_htbl);
 	new_ste = &new_htbl->ste_arr[new_idx];
@@ -403,7 +417,8 @@ static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
 	/* Write new table to HW */
 	info.type = CONNECT_MISS;
 	info.miss_icm_addr = nic_matcher->e_anchor->chunk->icm_addr;
-	dr_ste_set_formated_ste(dmn->info.caps.gvmi,
+	dr_ste_set_formated_ste(dmn->ste_ctx,
+				dmn->info.caps.gvmi,
 				nic_dmn,
 				new_htbl,
 				formated_ste,
@@ -450,18 +465,20 @@ static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
 		 * It is safe to operate dr_ste_set_hit_addr on the hw_ste here
 		 * (48B len) which works only on first 32B
 		 */
-		dr_ste_set_hit_addr(prev_htbl->ste_arr[0].hw_ste,
+		dr_ste_set_hit_addr(dmn->ste_ctx,
+				    prev_htbl->ste_arr[0].hw_ste,
 				    new_htbl->chunk->icm_addr,
 				    new_htbl->chunk->num_of_entries);
 
 		ste_to_update = &prev_htbl->ste_arr[0];
 	} else {
-		dr_ste_set_hit_addr_by_next_htbl(cur_htbl->pointing_ste->hw_ste,
+		dr_ste_set_hit_addr_by_next_htbl(dmn->ste_ctx,
+						 cur_htbl->pointing_ste->hw_ste,
 						 new_htbl);
 		ste_to_update = cur_htbl->pointing_ste;
 	}
 
-	dr_send_fill_and_append_ste_send_info(ste_to_update, DR_STE_SIZE_REDUCED,
+	dr_send_fill_and_append_ste_send_info(ste_to_update, DR_STE_SIZE_CTRL,
 					      0, ste_to_update->hw_ste, ste_info,
 					      update_list, false);
 
@@ -508,24 +525,26 @@ static struct dr_ste *dr_rule_handle_collision(struct mlx5dv_dr_matcher *matcher
 					       struct list_head *miss_list,
 					       struct list_head *send_list)
 {
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	struct dr_ste_ctx *ste_ctx = dmn->ste_ctx;
 	struct dr_ste_send_info *ste_info;
 	struct dr_ste *new_ste;
 
 	ste_info = calloc(1, sizeof(*ste_info));
 	if (!ste_info) {
-		dr_dbg(matcher->tbl->dmn, "Failed allocating ste_info\n");
+		dr_dbg(dmn, "Failed allocating ste_info\n");
 		errno = ENOMEM;
 		return NULL;
 	}
 
 	new_ste = dr_rule_create_collision_entry(matcher, nic_matcher, hw_ste, ste);
 	if (!new_ste) {
-		dr_dbg(matcher->tbl->dmn, "Failed creating collision entry\n");
+		dr_dbg(dmn, "Failed creating collision entry\n");
 		goto free_send_info;
 	}
 
-	if (dr_rule_append_to_miss_list(new_ste, miss_list, send_list)) {
-		dr_dbg(matcher->tbl->dmn, "Failed to update prev miss_list\n");
+	if (dr_rule_append_to_miss_list(ste_ctx, new_ste, miss_list, send_list)) {
+		dr_dbg(dmn, "Failed to update prev miss_list\n");
 		goto err_exit;
 	}
 
@@ -538,7 +557,8 @@ static struct dr_ste *dr_rule_handle_collision(struct mlx5dv_dr_matcher *matcher
 	return new_ste;
 
 err_exit:
-	dr_ste_free(new_ste, matcher, nic_matcher);
+	dr_htbl_put(new_ste->htbl);
+
 free_send_info:
 	free(ste_info);
 	return NULL;
@@ -583,32 +603,63 @@ free_action_members:
 	return errno;
 }
 
-/*
- * While the pointer of ste is no longer valid, like while moving ste to be
- * the first in the miss_list, and to be in the origin table,
- * all rule-members that are attached to this ste should update their ste member
- * to the new pointer
- */
-void dr_rule_update_rule_member(struct dr_ste *ste, struct dr_ste *new_ste)
+void dr_rule_set_last_member(struct dr_rule_rx_tx *nic_rule,
+			     struct dr_ste *ste,
+			     bool force)
 {
-	struct dr_rule_member *rule_mem;
+	/* Update rule member is usually done for the last STE or during rule
+	 * creation to recover from mid-creation failure (for this purpose the
+	 * force flag is used)
+	 */
+	if (ste->next_htbl && !force)
+		return;
 
-	list_for_each(&ste->rule_list, rule_mem, use_ste_list)
-		rule_mem->ste = new_ste;
+	/* Update is required since each rule keeps track of its last STE */
+	ste->rule_rx_tx = nic_rule;
+	nic_rule->last_rule_ste = ste;
+}
+
+static struct dr_ste *dr_rule_get_pointed_ste(struct dr_ste *curr_ste)
+{
+	struct dr_ste *first_ste;
+
+	first_ste = list_top(dr_ste_get_miss_list(curr_ste), struct dr_ste,
+			     miss_list_node);
+
+	return first_ste->htbl->pointing_ste;
+}
+
+void dr_rule_get_reverse_rule_members(struct dr_ste **ste_arr,
+				      struct dr_ste *curr_ste,
+				      int *num_of_stes)
+{
+	bool first = false;
+
+	*num_of_stes = 0;
+
+	if (curr_ste == NULL)
+		return;
+
+	/* Iterate from last to first */
+	while (!first) {
+		first = curr_ste->ste_chain_location == 1;
+		ste_arr[*num_of_stes] = curr_ste;
+		*num_of_stes += 1;
+		curr_ste = dr_rule_get_pointed_ste(curr_ste);
+	}
 }
 
 static void dr_rule_clean_rule_members(struct mlx5dv_dr_rule *rule,
 				       struct dr_rule_rx_tx *nic_rule)
 {
-	struct dr_rule_member *rule_mem;
-	struct dr_rule_member *tmp_mem;
+	struct dr_ste *ste_arr[DR_RULE_MAX_STES + DR_ACTION_MAX_STES];
+	struct dr_ste *curr_ste = nic_rule->last_rule_ste;
+	int i;
 
-	list_for_each_safe(&nic_rule->rule_members_list, rule_mem, tmp_mem, list) {
-		list_del(&rule_mem->list);
-		list_del(&rule_mem->use_ste_list);
-		dr_ste_put(rule_mem->ste, rule->matcher, nic_rule->nic_matcher);
-		free(rule_mem);
-	}
+	dr_rule_get_reverse_rule_members(ste_arr, curr_ste, &i);
+
+	while (i--)
+		dr_ste_put(ste_arr[i], rule->matcher, nic_rule->nic_matcher);
 }
 
 static uint16_t dr_get_bits_per_mask(uint16_t byte_mask)
@@ -645,25 +696,6 @@ static bool dr_rule_need_enlarge_hash(struct dr_ste_htbl *htbl,
 	return false;
 }
 
-static int dr_rule_add_member(struct dr_rule_rx_tx *nic_rule,
-			      struct dr_ste *ste)
-{
-	struct dr_rule_member *rule_mem;
-
-	rule_mem = calloc(1, sizeof(*rule_mem));
-	if (!rule_mem) {
-		errno = ENOMEM;
-		return errno;
-	}
-
-	rule_mem->ste = ste;
-	list_add_tail(&nic_rule->rule_members_list, &rule_mem->list);
-
-	list_add_tail(&ste->rule_list, &rule_mem->use_ste_list);
-
-	return 0;
-}
-
 static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 				      struct dr_rule_rx_tx *nic_rule,
 				      struct list_head *send_ste_list,
@@ -675,6 +707,7 @@ static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 	struct dr_ste_send_info *ste_info_arr[DR_ACTION_MAX_STES];
 	uint8_t num_of_builders = nic_matcher->num_of_builders;
 	struct mlx5dv_dr_matcher *matcher = rule->matcher;
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
 	uint8_t *curr_hw_ste, *prev_hw_ste;
 	struct dr_ste *action_ste;
 	int i, k, ret;
@@ -684,8 +717,10 @@ static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 	 * 2. num_of_builders is less then new_hw_ste_arr_sz, new ste was added
 	 *    to support the action.
 	 */
-	if (num_of_builders == new_hw_ste_arr_sz)
+	if (num_of_builders == new_hw_ste_arr_sz) {
+		last_ste->next_htbl = NULL;
 		return 0;
+	}
 
 	for (i = num_of_builders, k = 0; i < new_hw_ste_arr_sz; i++, k++) {
 		curr_hw_ste = hw_ste_arr + i * DR_STE_SIZE;
@@ -698,25 +733,29 @@ static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 
 		dr_ste_get(action_ste);
 
+		action_ste->htbl->pointing_ste = last_ste;
+		last_ste->next_htbl = action_ste->htbl;
+		last_ste = action_ste;
+
 		/* While free ste we go over the miss list, so add this ste to the list */
 		list_add_tail(dr_ste_get_miss_list(action_ste),
 			      &action_ste->miss_list_node);
 
 		ste_info_arr[k] = calloc(1, sizeof(struct dr_ste_send_info));
 		if (!ste_info_arr[k]) {
-			dr_dbg(matcher->tbl->dmn, "Failed allocate ste_info, k: %d\n", k);
+			dr_dbg(dmn, "Failed allocate ste_info, k: %d\n", k);
 			errno = ENOMEM;
 			ret = errno;
 			goto err_exit;
 		}
 
 		/* Point current ste to the new action */
-		dr_ste_set_hit_addr_by_next_htbl(prev_hw_ste, action_ste->htbl);
-		ret = dr_rule_add_member(nic_rule, action_ste);
-		if (ret) {
-			dr_dbg(matcher->tbl->dmn, "Failed adding rule member\n");
-			goto free_ste_info;
-		}
+		dr_ste_set_hit_addr_by_next_htbl(dmn->ste_ctx,
+						 prev_hw_ste,
+						 action_ste->htbl);
+
+		dr_rule_set_last_member(nic_rule, action_ste, true);
+
 		dr_send_fill_and_append_ste_send_info(action_ste, DR_STE_SIZE, 0,
 						      curr_hw_ste,
 						      ste_info_arr[k],
@@ -725,8 +764,6 @@ static int dr_rule_handle_action_stes(struct mlx5dv_dr_rule *rule,
 
 	return 0;
 
-free_ste_info:
-	free(ste_info_arr[k]);
 err_exit:
 	dr_ste_put(action_ste, matcher, nic_matcher);
 	return ret;
@@ -741,6 +778,7 @@ static int dr_rule_handle_empty_entry(struct mlx5dv_dr_matcher *matcher,
 				      struct list_head *miss_list,
 				      struct list_head *send_list)
 {
+	struct dr_ste_ctx *ste_ctx = matcher->tbl->dmn->ste_ctx;
 	struct dr_ste_send_info *ste_info;
 
 	/* Take ref on table, only on first time this ste is used */
@@ -749,7 +787,7 @@ static int dr_rule_handle_empty_entry(struct mlx5dv_dr_matcher *matcher,
 	/* new entry -> new branch */
 	list_add_tail(miss_list, &ste->miss_list_node);
 
-	dr_ste_set_miss_addr(hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
+	dr_ste_set_miss_addr(ste_ctx, hw_ste, nic_matcher->e_anchor->chunk->icm_addr);
 
 	ste->ste_chain_location = ste_location;
 
@@ -895,8 +933,10 @@ static bool dr_rule_verify(struct mlx5dv_dr_matcher *matcher,
 	size_t value_size = value->match_sz;
 	uint32_t s_idx, e_idx;
 
-	if (!value_size ||
-	    (value_size > DEVX_ST_SZ_BYTES(dr_match_param) ||
+	if (!value_size)
+		return true;
+
+	if ((value_size > DEVX_ST_SZ_BYTES(dr_match_param) ||
 	     (value_size % sizeof(uint32_t)))) {
 		dr_dbg(dmn, "Rule parameters length is incorrect\n");
 		errno = EINVAL;
@@ -954,13 +994,25 @@ static bool dr_rule_verify(struct mlx5dv_dr_matcher *matcher,
 			return false;
 		}
 	}
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC4) {
+		s_idx = offsetof(struct dr_match_param, misc4);
+		e_idx = min(s_idx + sizeof(param->misc4), value_size);
+
+		if (!dr_rule_cmp_value_to_mask(mask_p, param_p, s_idx, e_idx)) {
+			dr_dbg(dmn, "Rule misc4 parameters contains a value not specified by mask\n");
+			return false;
+		}
+	}
 	return true;
 }
 
 static int dr_rule_destroy_rule_nic(struct mlx5dv_dr_rule *rule,
 				    struct dr_rule_rx_tx *nic_rule)
 {
+	dr_domain_nic_lock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
 	dr_rule_clean_rule_members(rule, nic_rule);
+	dr_domain_nic_unlock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
 	return 0;
 }
 
@@ -975,6 +1027,10 @@ static int dr_rule_destroy_rule(struct mlx5dv_dr_rule *rule)
 {
 	struct mlx5dv_dr_domain *dmn = rule->matcher->tbl->dmn;
 
+	dr_domain_lock(dmn);
+	list_del(&rule->rule_list);
+	dr_domain_unlock(dmn);
+
 	switch (dmn->type) {
 	case MLX5DV_DR_DOMAIN_TYPE_NIC_RX:
 		dr_rule_destroy_rule_nic(rule, &rule->rx);
@@ -986,12 +1042,13 @@ static int dr_rule_destroy_rule(struct mlx5dv_dr_rule *rule)
 		dr_rule_destroy_rule_fdb(rule);
 		break;
 	default:
+		assert(false);
 		errno = EINVAL;
 		return errno;
 	}
 
 	dr_rule_remove_action_members(rule);
-	list_del(&rule->rule_list);
+
 	free(rule);
 	return 0;
 }
@@ -1049,22 +1106,22 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	struct dr_ste *ste = NULL; /* Fix compilation warning */
 	int ret, i;
 
-	list_head_init(&nic_rule->rule_members_list);
-
 	if (dr_rule_skip(dmn->type, nic_dmn->ste_type, &matcher->mask, param))
 		return 0;
 
 	/* Set the tag values inside the ste array */
 	ret = dr_ste_build_ste_arr(matcher, nic_matcher, param, hw_ste_arr);
 	if (ret)
-		goto out_err;
+		return ret;
 
 	/* Set the actions values/addresses inside the ste array */
 	ret = dr_actions_build_ste_arr(matcher, nic_matcher, actions,
 				       num_actions, hw_ste_arr,
 				       &new_hw_ste_arr_sz);
 	if (ret)
-		goto out_err;
+		return ret;
+
+	dr_domain_nic_lock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
 
 	cur_htbl = nic_matcher->s_htbl;
 
@@ -1091,15 +1148,8 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 		}
 
 		cur_htbl = ste->next_htbl;
-
-		/* Keep all STEs in the rule struct */
-		ret = dr_rule_add_member(nic_rule, ste);
-		if (ret) {
-			dr_dbg(dmn, "Failed adding rule member index %d\n", i);
-			goto free_ste;
-		}
-
 		dr_ste_get(ste);
+		dr_rule_set_last_member(nic_rule, ste, true);
 	}
 
 	/* Connect actions */
@@ -1118,10 +1168,8 @@ dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
 	if (htbl)
 		dr_htbl_put(htbl);
 
-	return 0;
+	goto out_unlock;
 
-free_ste:
-	dr_ste_put(ste, matcher, nic_matcher);
 free_rule:
 	dr_rule_clean_rule_members(rule, nic_rule);
 	/* Clean all ste_info's */
@@ -1129,7 +1177,8 @@ free_rule:
 		list_del(&ste_info->send_list);
 		free(ste_info);
 	}
-out_err:
+out_unlock:
+	dr_domain_nic_unlock(nic_rule->nic_matcher->nic_tbl->nic_dmn);
 	return ret;
 }
 
@@ -1219,7 +1268,10 @@ dr_rule_create_rule(struct mlx5dv_dr_matcher *matcher,
 	if (ret)
 		goto remove_action_members;
 
+	dr_domain_lock(dmn);
 	list_add_tail(&matcher->rule_list, &rule->rule_list);
+	dr_domain_unlock(dmn);
+
 	return rule;
 
 remove_action_members:
@@ -1301,7 +1353,6 @@ struct mlx5dv_dr_rule *mlx5dv_dr_rule_create(struct mlx5dv_dr_matcher *matcher,
 {
 	struct mlx5dv_dr_rule *rule;
 
-	pthread_mutex_lock(&matcher->tbl->dmn->mutex);
 	atomic_fetch_add(&matcher->refcount, 1);
 
 	if (dr_is_root_table(matcher->tbl))
@@ -1312,8 +1363,6 @@ struct mlx5dv_dr_rule *mlx5dv_dr_rule_create(struct mlx5dv_dr_matcher *matcher,
 	if (!rule)
 		atomic_fetch_sub(&matcher->refcount, 1);
 
-	pthread_mutex_unlock(&matcher->tbl->dmn->mutex);
-
 	return rule;
 }
 
@@ -1323,15 +1372,11 @@ int mlx5dv_dr_rule_destroy(struct mlx5dv_dr_rule *rule)
 	struct mlx5dv_dr_table *tbl = rule->matcher->tbl;
 	int ret;
 
-	pthread_mutex_lock(&tbl->dmn->mutex);
-
 	if (dr_is_root_table(tbl))
 		ret = dr_rule_destroy_rule_root(rule);
 	else
 		ret = dr_rule_destroy_rule(rule);
 
-	pthread_mutex_unlock(&tbl->dmn->mutex);
-
 	if (!ret)
 		atomic_fetch_sub(&matcher->refcount, 1);
 	return ret;
