@@ -43,6 +43,7 @@
 #include <sys/param.h>
 
 #include <util/symver.h>
+#include <rdma/mlx5_user_ioctl_cmds.h>
 
 #include "mlx5.h"
 #include "mlx5-abi.h"
@@ -51,10 +52,6 @@
 
 static void mlx5_free_context(struct ibv_context *ibctx);
 
-#ifndef PCI_VENDOR_ID_MELLANOX
-#define PCI_VENDOR_ID_MELLANOX			0x15b3
-#endif
-
 #ifndef CPU_OR
 #define CPU_OR(x, y, z) do {} while (0)
 #endif
@@ -80,10 +77,12 @@ static const struct verbs_match_ent hca_table[] = {
 	HCA(MELLANOX, 0x101c),	/* ConnectX-6 VF */
 	HCA(MELLANOX, 0x101d),	/* ConnectX-6 DX */
 	HCA(MELLANOX, 0x101e),	/* ConnectX family mlx5Gen Virtual Function */
+	HCA(MELLANOX, 0x101f),	/* ConnectX-6 LX */
 	HCA(MELLANOX, 0x1021),  /* ConnectX-7 */
 	HCA(MELLANOX, 0xa2d2),	/* BlueField integrated ConnectX-5 network controller */
 	HCA(MELLANOX, 0xa2d3),	/* BlueField integrated ConnectX-5 network controller VF */
 	HCA(MELLANOX, 0xa2d6),  /* BlueField-2 integrated ConnectX-6 Dx network controller */
+	HCA(MELLANOX, 0xa2dc),  /* BlueField-3 integrated ConnectX-7 network controller */
 	{}
 };
 
@@ -91,7 +90,6 @@ uint32_t mlx5_debug_mask = 0;
 int mlx5_freeze_on_error_cqe;
 
 static const struct verbs_context_ops mlx5_ctx_common_ops = {
-	.query_device  = mlx5_query_device,
 	.query_port    = mlx5_query_port,
 	.alloc_pd      = mlx5_alloc_pd,
 	.async_event   = mlx5_async_event,
@@ -146,6 +144,8 @@ static const struct verbs_context_ops mlx5_ctx_common_ops = {
 	.destroy_wq = mlx5_destroy_wq,
 	.free_dm = mlx5_free_dm,
 	.get_srq_num = mlx5_get_srq_num,
+	.import_mr = mlx5_import_mr,
+	.import_pd = mlx5_import_pd,
 	.modify_cq = mlx5_modify_cq,
 	.modify_flow_action_esp = mlx5_modify_flow_action_esp,
 	.modify_qp_rate_limit = mlx5_modify_qp_rate_limit,
@@ -154,11 +154,15 @@ static const struct verbs_context_ops mlx5_ctx_common_ops = {
 	.open_xrcd = mlx5_open_xrcd,
 	.post_srq_ops = mlx5_post_srq_ops,
 	.query_device_ex = mlx5_query_device_ex,
+	.query_ece = mlx5_query_ece,
 	.query_rt_values = mlx5_query_rt_values,
 	.read_counters = mlx5_read_counters,
 	.reg_dm_mr = mlx5_reg_dm_mr,
 	.alloc_null_mr = mlx5_alloc_null_mr,
 	.free_context = mlx5_free_context,
+	.set_ece = mlx5_set_ece,
+	.unimport_mr = mlx5_unimport_mr,
+	.unimport_pd = mlx5_unimport_pd,
 };
 
 static const struct verbs_context_ops mlx5_ctx_cqev1_ops = {
@@ -320,7 +324,8 @@ man cpuset
   The first "1" is for bit 64, the second for bit 32, the third for bit 16, the fourth for bit 8, the fifth for
   bit 4, and the "7" is for bits 2, 1, and 0.
 */
-static void mlx5_local_cpu_set(struct ibv_device *ibdev, cpu_set_t *cpu_set)
+static void mlx5_local_cpu_set(struct ibv_device *ibdev, struct mlx5_context *mctx,
+			       cpu_set_t *cpu_set)
 {
 	char *p, buf[1024] = {};
 	char *env_value;
@@ -339,11 +344,11 @@ static void mlx5_local_cpu_set(struct ibv_device *ibdev, cpu_set_t *cpu_set)
 
 		fp = fopen(fname, "r");
 		if (!fp) {
-			fprintf(stderr, PFX "Warning: can not get local cpu set: failed to open %s\n", fname);
+			mlx5_err(mctx->dbg_fp, PFX "Warning: can not get local cpu set: failed to open %s\n", fname);
 			return;
 		}
 		if (!fgets(buf, sizeof(buf), fp)) {
-			fprintf(stderr, PFX "Warning: can not get local cpu set: failed to read cpu mask\n");
+			mlx5_err(mctx->dbg_fp, PFX "Warning: can not get local cpu set: failed to read cpu mask\n");
 			fclose(fp);
 			return;
 		}
@@ -378,7 +383,7 @@ static void mlx5_local_cpu_set(struct ibv_device *ibdev, cpu_set_t *cpu_set)
 	} while (i < CPU_SETSIZE);
 }
 
-static int mlx5_enable_sandy_bridge_fix(struct ibv_device *ibdev)
+static int mlx5_enable_sandy_bridge_fix(struct ibv_device *ibdev, struct mlx5_context *mctx)
 {
 	cpu_set_t my_cpus, dev_local_cpus, result_set;
 	int stall_enable;
@@ -403,14 +408,14 @@ static int mlx5_enable_sandy_bridge_fix(struct ibv_device *ibdev)
 	ret = sched_getaffinity(0, sizeof(my_cpus), &my_cpus);
 	if (ret == -1) {
 		if (errno == EINVAL)
-			fprintf(stderr, PFX "Warning: my cpu set is too small\n");
+			mlx5_err(mctx->dbg_fp, PFX "Warning: my cpu set is too small\n");
 		else
-			fprintf(stderr, PFX "Warning: failed to get my cpu set\n");
+			mlx5_err(mctx->dbg_fp, PFX "Warning: failed to get my cpu set\n");
 		goto out;
 	}
 
 	/* get device local cpu set */
-	mlx5_local_cpu_set(ibdev, &dev_local_cpus);
+	mlx5_local_cpu_set(ibdev, mctx, &dev_local_cpus);
 
 	/* check if my cpu set is in dev cpu */
 	CPU_OR(&result_set, &my_cpus, &dev_local_cpus);
@@ -430,7 +435,7 @@ static void mlx5_read_env(struct ibv_device *ibdev, struct mlx5_context *ctx)
 		ctx->stall_enable = (strcmp(env_value, "0")) ? 1 : 0;
 	else
 		/* autodetect if we need to do cq polling */
-		ctx->stall_enable = mlx5_enable_sandy_bridge_fix(ibdev);
+		ctx->stall_enable = mlx5_enable_sandy_bridge_fix(ibdev, ctx);
 
 	env_value = getenv("MLX5_STALL_NUM_LOOP");
 	if (env_value)
@@ -487,17 +492,22 @@ static int get_total_uuars(int page_size)
 static void open_debug_file(struct mlx5_context *ctx)
 {
 	char *env;
+	FILE *default_dbg_fp = NULL;
+
+#ifdef MLX5_DEBUG
+	default_dbg_fp = stderr;
+#endif
 
 	env = getenv("MLX5_DEBUG_FILE");
 	if (!env) {
-		ctx->dbg_fp = stderr;
+		ctx->dbg_fp = default_dbg_fp;
 		return;
 	}
 
 	ctx->dbg_fp = fopen(env, "aw+");
 	if (!ctx->dbg_fp) {
-		fprintf(stderr, "Failed opening debug file %s, using stderr\n", env);
-		ctx->dbg_fp = stderr;
+		ctx->dbg_fp = default_dbg_fp;
+		mlx5_err(ctx->dbg_fp, "Failed opening debug file %s\n", env);
 		return;
 	}
 }
@@ -651,9 +661,9 @@ static int mlx5_map_internal_clock(struct mlx5_device *mdev,
 			      mdev->page_size * offset);
 
 	if (hca_clock_page == MAP_FAILED) {
-		fprintf(stderr, PFX
-			"Warning: Timestamp available,\n"
-			"but failed to mmap() hca core clock page.\n");
+		mlx5_err(context->dbg_fp, PFX
+			 "Warning: Timestamp available,\n"
+			 "but failed to mmap() hca core clock page.\n");
 		return -1;
 	}
 
@@ -796,6 +806,14 @@ int mlx5dv_query_device(struct ibv_context *ctx_in,
 		}
 	}
 
+	if (attrs_out->comp_mask & MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS) {
+		if (mctx->entropy_caps.num_lag_ports) {
+			attrs_out->num_lag_ports =
+				mctx->entropy_caps.num_lag_ports;
+			comp_mask_out |= MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS;
+		}
+	}
+
 	attrs_out->comp_mask = comp_mask_out;
 
 	return 0;
@@ -865,11 +883,11 @@ static int mlx5dv_get_cq(struct ibv_cq *cq_in,
 
 	cq_out->comp_mask = 0;
 	cq_out->cqn       = mcq->cqn;
-	cq_out->cqe_cnt   = mcq->ibv_cq.cqe + 1;
+	cq_out->cqe_cnt   = mcq->verbs_cq.cq.cqe + 1;
 	cq_out->cqe_size  = mcq->cqe_sz;
 	cq_out->buf       = mcq->active_buf->buf;
 	cq_out->dbrec     = mcq->dbrec;
-	cq_out->cq_uar	  = mctx->uar[0].reg;
+	cq_out->cq_uar	  = mctx->cq_uar_reg;
 
 	mcq->flags	 |= MLX5_CQ_FLAGS_DV_OWNED;
 
@@ -969,6 +987,576 @@ static int mlx5dv_get_pd(struct ibv_pd *pd_in,
 	return 0;
 }
 
+static int query_lag(struct ibv_context *ctx, uint8_t *lag_state,
+		     uint8_t *tx_remap_affinity_1,
+		     uint8_t *tx_remap_affinity_2)
+{
+	uint32_t out_lag[DEVX_ST_SZ_DW(query_lag_out)] = {};
+	uint32_t in_lag[DEVX_ST_SZ_DW(query_lag_in)] = {};
+	int ret;
+
+	DEVX_SET(query_lag_in, in_lag, opcode, MLX5_CMD_OP_QUERY_LAG);
+	ret = mlx5dv_devx_general_cmd(ctx, in_lag, sizeof(in_lag), out_lag,
+				      sizeof(out_lag));
+	if (ret)
+		return ret;
+
+	*lag_state = DEVX_GET(query_lag_out, out_lag, ctx.lag_state);
+	if (tx_remap_affinity_1)
+		*tx_remap_affinity_1 = DEVX_GET(query_lag_out, out_lag,
+						ctx.tx_remap_affinity_1);
+	if (tx_remap_affinity_2)
+		*tx_remap_affinity_2 = DEVX_GET(query_lag_out, out_lag,
+						ctx.tx_remap_affinity_2);
+
+	return 0;
+}
+
+static bool lag_operation_supported(struct ibv_qp *qp)
+{
+	struct mlx5_context *mctx = to_mctx(qp->context);
+	struct mlx5_qp *mqp = to_mqp(qp);
+
+	if (!is_mlx5_dev(qp->context->device) ||
+	    (mctx->entropy_caps.num_lag_ports <= 1))
+		return false;
+
+	if ((qp->qp_type == IBV_QPT_RC) ||
+	    (qp->qp_type == IBV_QPT_UD) ||
+	    (qp->qp_type == IBV_QPT_UC) ||
+	    (qp->qp_type == IBV_QPT_RAW_PACKET) ||
+	    (qp->qp_type == IBV_QPT_XRC_SEND) ||
+	    ((qp->qp_type == IBV_QPT_DRIVER) &&
+	     (mqp->dc_type == MLX5DV_DCTYPE_DCI)))
+		return true;
+
+	return false;
+}
+
+
+int mlx5dv_query_qp_lag_port(struct ibv_qp *qp, uint8_t *port_num,
+			     uint8_t *active_port_num)
+{
+	uint8_t lag_state, tx_remap_affinity_1, tx_remap_affinity_2;
+	uint32_t in_tis[DEVX_ST_SZ_DW(query_tis_in)] = {};
+	uint32_t out_tis[DEVX_ST_SZ_DW(query_tis_out)] = {};
+	uint32_t in_qp[DEVX_ST_SZ_DW(query_qp_in)] = {};
+	uint32_t out_qp[DEVX_ST_SZ_DW(query_qp_out)] = {};
+	struct mlx5_context *mctx = to_mctx(qp->context);
+	struct mlx5_qp *mqp = to_mqp(qp);
+	int ret;
+
+	if (!lag_operation_supported(qp))
+		return EOPNOTSUPP;
+
+	ret = query_lag(qp->context, &lag_state,
+			&tx_remap_affinity_1, &tx_remap_affinity_2);
+	if (ret)
+		return ret;
+
+	if (!lag_state && !mctx->entropy_caps.lag_tx_port_affinity)
+		return EOPNOTSUPP;
+
+	switch (qp->qp_type) {
+	case IBV_QPT_RAW_PACKET:
+		DEVX_SET(query_tis_in, in_tis, opcode, MLX5_CMD_OP_QUERY_TIS);
+		DEVX_SET(query_tis_in, in_tis, tisn, mqp->tisn);
+		ret = mlx5dv_devx_qp_query(qp, in_tis, sizeof(in_tis), out_tis,
+					   sizeof(out_tis));
+		if (ret)
+			return ret;
+
+		*port_num = DEVX_GET(query_tis_out, out_tis,
+				     tis_context.lag_tx_port_affinity);
+		break;
+
+	default:
+		DEVX_SET(query_qp_in, in_qp, opcode, MLX5_CMD_OP_QUERY_QP);
+		DEVX_SET(query_qp_in, in_qp, qpn, qp->qp_num);
+		ret = mlx5dv_devx_qp_query(qp, in_qp, sizeof(in_qp), out_qp,
+					   sizeof(out_qp));
+		if (ret)
+			return ret;
+
+		*port_num = DEVX_GET(query_qp_out, out_qp,
+				     qpc.lag_tx_port_affinity);
+		break;
+	}
+
+	switch (*port_num) {
+	case 1:
+		*active_port_num = tx_remap_affinity_1;
+		break;
+
+	case 2:
+		*active_port_num = tx_remap_affinity_2;
+		break;
+
+	default:
+		return EOPNOTSUPP;
+	}
+
+	return 0;
+}
+
+static int modify_tis_lag_port(struct ibv_qp *qp, uint8_t port_num)
+{
+	uint32_t out[DEVX_ST_SZ_DW(modify_tis_out)] = {};
+	uint32_t in[DEVX_ST_SZ_DW(modify_tis_in)] = {};
+	struct mlx5_qp *mqp = to_mqp(qp);
+
+	DEVX_SET(modify_tis_in, in, opcode, MLX5_CMD_OP_MODIFY_TIS);
+	DEVX_SET(modify_tis_in, in, tisn, mqp->tisn);
+	DEVX_SET(modify_tis_in, in, bitmask.lag_tx_port_affinity, 1);
+	DEVX_SET(modify_tis_in, in, ctx.lag_tx_port_affinity, port_num);
+	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
+}
+
+static int modify_qp_lag_port(struct ibv_qp *qp, uint8_t port_num)
+{
+	uint32_t out[DEVX_ST_SZ_DW(rts2rts_qp_out)] = {};
+	uint32_t in[DEVX_ST_SZ_DW(rts2rts_qp_in)] = {};
+	struct mlx5_context *mctx = to_mctx(qp->context);
+
+	if (!mctx->entropy_caps.rts2rts_lag_tx_port_affinity ||
+	    qp->state != IBV_QPS_RTS)
+		return EOPNOTSUPP;
+
+	DEVX_SET(rts2rts_qp_in, in, opcode, MLX5_CMD_OP_RTS2RTS_QP);
+	DEVX_SET(rts2rts_qp_in, in, qpn, qp->qp_num);
+	DEVX_SET(rts2rts_qp_in, in, opt_param_mask,
+		 MLX5_QPC_OPT_MASK_RTS2RTS_LAG_TX_PORT_AFFINITY);
+	DEVX_SET(rts2rts_qp_in, in, qpc.lag_tx_port_affinity, port_num);
+	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
+}
+
+int mlx5dv_modify_qp_lag_port(struct ibv_qp *qp, uint8_t port_num)
+{
+	uint8_t curr_configured, curr_active;
+	struct mlx5_qp *mqp = to_mqp(qp);
+	int ret;
+
+	/* Query lag port to see if we are at all in lag mode, otherwise FW
+	 * might return success and ignore the modification.
+	 */
+	ret = mlx5dv_query_qp_lag_port(qp, &curr_configured, &curr_active);
+	if (ret)
+		return ret;
+
+	switch (qp->qp_type) {
+	case IBV_QPT_RAW_PACKET:
+		return modify_tis_lag_port(qp, port_num);
+
+	case IBV_QPT_DRIVER:
+		if (mqp->dc_type != MLX5DV_DCTYPE_DCI)
+			return EOPNOTSUPP;
+		SWITCH_FALLTHROUGH;
+	case IBV_QPT_RC:
+	case IBV_QPT_UD:
+	case IBV_QPT_UC:
+		return modify_qp_lag_port(qp, port_num);
+
+	default:
+		return EOPNOTSUPP;
+	}
+}
+
+int mlx5dv_modify_qp_udp_sport(struct ibv_qp *qp, uint16_t udp_sport)
+{
+	uint32_t in[DEVX_ST_SZ_DW(rts2rts_qp_in)] = {};
+	uint32_t out[DEVX_ST_SZ_DW(rts2rts_qp_out)] = {};
+	struct mlx5_context *mctx = to_mctx(qp->context);
+
+	if (!is_mlx5_dev(qp->context->device))
+		return EOPNOTSUPP;
+
+	switch (qp->qp_type) {
+	case IBV_QPT_RC:
+	case IBV_QPT_UC:
+		if (qp->state != IBV_QPS_RTS ||
+		    !mctx->entropy_caps.rts2rts_qp_udp_sport)
+			return EOPNOTSUPP;
+		break;
+	default:
+		return EOPNOTSUPP;
+	}
+	DEVX_SET(rts2rts_qp_in, in, opcode, MLX5_CMD_OP_RTS2RTS_QP);
+	DEVX_SET(rts2rts_qp_in, in, qpn, qp->qp_num);
+	DEVX_SET64(rts2rts_qp_in, in, opt_param_mask_95_32,
+		   MLX5_QPC_OPT_MASK_32_UDP_SPORT);
+	DEVX_SET(rts2rts_qp_in, in, qpc.primary_address_path.udp_sport,
+		 udp_sport);
+
+	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out,
+				     sizeof(out));
+}
+
+static bool sched_supported(struct ibv_context *ctx)
+{
+	struct mlx5_qos_caps *qc = &to_mctx(ctx)->qos_caps;
+
+	return (qc->qos &&
+		(qc->nic_element_type & ELEMENT_TYPE_CAP_MASK_TASR) &&
+		(qc->nic_element_type & ELEMENT_TYPE_CAP_MASK_QUEUE_GROUP) &&
+		(qc->nic_tsar_type & TSAR_TYPE_CAP_MASK_DWRR));
+}
+
+static struct mlx5dv_devx_obj *
+mlx5dv_sched_nic_create(struct ibv_context *ctx,
+			const struct mlx5dv_sched_attr *sched_attr,
+			int elem_type)
+{
+	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {};
+	uint32_t in[DEVX_ST_SZ_DW(create_sched_elem_in)] = {};
+	uint32_t parent_id;
+	void *attr;
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, hdr);
+	DEVX_SET(general_obj_in_cmd_hdr,
+		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
+	DEVX_SET(general_obj_in_cmd_hdr,
+		 attr, obj_type, MLX5_OBJ_TYPE_SCHEDULING_ELEMENT);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, sched_elem);
+	DEVX_SET64(sched_elem, attr, modify_field_select, sched_attr->flags);
+	DEVX_SET(sched_elem, attr,
+		 scheduling_hierarchy, MLX5_SCHED_HIERARCHY_NIC);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, sched_elem.sched_context);
+	DEVX_SET(sched_context, attr, element_type, elem_type);
+
+	parent_id = sched_attr->parent ? sched_attr->parent->obj->object_id : 0;
+	DEVX_SET(sched_context, attr, parent_element_id, parent_id);
+	if (sched_attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)
+		DEVX_SET(sched_context, attr, bw_share, sched_attr->bw_share);
+	if (sched_attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)
+		DEVX_SET(sched_context, attr,
+			 max_average_bw, sched_attr->max_avg_bw);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in,
+			    sched_elem.sched_context.sched_elem_attr);
+	DEVX_SET(sched_elem_attr_tsar, attr, tsar_type,
+		 MLX5_SCHED_TSAR_TYPE_DWRR);
+
+	return mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
+}
+
+static int
+mlx5dv_sched_nic_modify(struct mlx5dv_devx_obj *obj,
+			const struct mlx5dv_sched_attr *sched_attr,
+			int elem_type)
+{
+	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {};
+	uint32_t in[DEVX_ST_SZ_DW(create_sched_elem_in)] = {};
+	void *attr;
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, hdr);
+	DEVX_SET(general_obj_in_cmd_hdr,
+		 attr, opcode, MLX5_CMD_OP_MODIFY_GENERAL_OBJECT);
+	DEVX_SET(general_obj_in_cmd_hdr,
+		 attr, obj_type, MLX5_OBJ_TYPE_SCHEDULING_ELEMENT);
+	DEVX_SET(general_obj_in_cmd_hdr, in, obj_id, obj->object_id);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, sched_elem);
+	DEVX_SET64(sched_elem, attr, modify_field_select, sched_attr->flags);
+	DEVX_SET(sched_elem, attr,
+		 scheduling_hierarchy, MLX5_SCHED_HIERARCHY_NIC);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in, sched_elem.sched_context);
+	DEVX_SET(sched_context, attr, element_type, elem_type);
+	if (sched_attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)
+		DEVX_SET(sched_context, attr, bw_share, sched_attr->bw_share);
+	if (sched_attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)
+		DEVX_SET(sched_context, attr,
+			 max_average_bw, sched_attr->max_avg_bw);
+
+	attr = DEVX_ADDR_OF(create_sched_elem_in, in,
+			    sched_elem.sched_context.sched_elem_attr);
+	DEVX_SET(sched_elem_attr_tsar, attr, tsar_type,
+		 MLX5_SCHED_TSAR_TYPE_DWRR);
+
+	return mlx5dv_devx_obj_modify(obj, in, sizeof(in), out, sizeof(out));
+}
+
+#define MLX5DV_SCHED_ELEM_ATTR_ALL_FLAGS \
+	(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE |	\
+	 MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)
+
+static bool attr_supported(struct ibv_context *ctx,
+			   const struct mlx5dv_sched_attr *attr)
+{
+	struct mlx5_qos_caps *qc = &to_mctx(ctx)->qos_caps;
+
+	if ((attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE) &&
+	    !qc->nic_bw_share)
+		return false;
+	if ((attr->flags & MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW) &&
+	    !qc->nic_rate_limit)
+		return false;
+
+	return true;
+}
+
+static bool sched_attr_valid(const struct mlx5dv_sched_attr *attr, bool node)
+{
+	if (!attr || attr->comp_mask ||
+	    !check_comp_mask(attr->flags, MLX5DV_SCHED_ELEM_ATTR_ALL_FLAGS))
+		return false;
+	if (node && (!attr->parent && attr->flags))
+		return false;
+	if (!node && !attr->parent)
+		return false;
+
+	return true;
+}
+
+struct mlx5dv_sched_node *
+mlx5dv_sched_node_create(struct ibv_context *ctx,
+			 const struct mlx5dv_sched_attr *attr)
+{
+	struct mlx5dv_sched_node *node;
+	struct mlx5dv_devx_obj *obj;
+
+	if (!is_mlx5_dev(ctx->device)) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	if (!sched_attr_valid(attr, true)) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (!sched_supported(ctx) || !attr_supported(ctx, attr)) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	node = calloc(1, sizeof(*node));
+	if (!node) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	obj = mlx5dv_sched_nic_create(ctx, attr, MLX5_SCHED_ELEM_TYPE_TSAR);
+	if (!obj)
+		goto err_sched_nic_create;
+
+	node->obj = obj;
+	node->parent = attr->parent;
+	return node;
+
+err_sched_nic_create:
+	free(node);
+	return NULL;
+}
+
+struct mlx5dv_sched_leaf *
+mlx5dv_sched_leaf_create(struct ibv_context *ctx,
+			 const struct mlx5dv_sched_attr *attr)
+{
+	struct mlx5dv_sched_leaf *leaf;
+	struct mlx5dv_devx_obj *obj;
+
+	if (!sched_attr_valid(attr, false)) {
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (!attr_supported(ctx, attr)) {
+		errno = EOPNOTSUPP;
+		return NULL;
+	}
+
+	leaf = calloc(1, sizeof(*leaf));
+	if (!leaf) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	obj = mlx5dv_sched_nic_create(ctx, attr,
+				      MLX5_SCHED_ELEM_TYPE_QUEUE_GROUP);
+	if (!obj)
+		goto err_sched_nic_create;
+
+	leaf->obj = obj;
+	leaf->parent = attr->parent;
+	return leaf;
+
+err_sched_nic_create:
+	free(leaf);
+	return NULL;
+}
+
+int mlx5dv_sched_node_modify(struct mlx5dv_sched_node *node,
+			     const struct mlx5dv_sched_attr *attr)
+{
+	if (!node || !sched_attr_valid(attr, true)) {
+		errno = EINVAL;
+		return errno;
+	}
+
+	if (!attr_supported(node->obj->context, attr)) {
+		errno = EOPNOTSUPP;
+		return errno;
+	}
+
+	return mlx5dv_sched_nic_modify(node->obj, attr,
+				       MLX5_SCHED_ELEM_TYPE_TSAR);
+}
+
+int mlx5dv_sched_leaf_modify(struct mlx5dv_sched_leaf *leaf,
+			     const struct mlx5dv_sched_attr *attr)
+{
+	if (!leaf || !sched_attr_valid(attr, false)) {
+		errno = EINVAL;
+		return errno;
+	}
+
+	if (!attr_supported(leaf->obj->context, attr)) {
+		errno = EOPNOTSUPP;
+		return errno;
+	}
+
+	return mlx5dv_sched_nic_modify(leaf->obj, attr,
+				       MLX5_SCHED_ELEM_TYPE_QUEUE_GROUP);
+}
+
+int mlx5dv_sched_node_destroy(struct mlx5dv_sched_node *node)
+{
+	int ret;
+
+	ret = mlx5dv_devx_obj_destroy(node->obj);
+	if (ret)
+		return ret;
+
+	free(node);
+	return 0;
+}
+
+int mlx5dv_sched_leaf_destroy(struct mlx5dv_sched_leaf *leaf)
+{
+	int ret;
+
+	ret = mlx5dv_devx_obj_destroy(leaf->obj);
+	if (ret)
+		return ret;
+
+	free(leaf);
+	return 0;
+}
+
+static int modify_ib_qp_sched_elem_init(struct ibv_qp *qp,
+					uint32_t req_id, uint32_t resp_id)
+{
+	uint64_t mask = MLX5_QPC_OPT_MASK_32_QOS_QUEUE_GROUP_ID;
+	uint32_t in[DEVX_ST_SZ_DW(init2init_qp_in)] = {};
+	uint32_t out[DEVX_ST_SZ_DW(init2init_qp_out)] = {};
+	void *qpce = DEVX_ADDR_OF(init2init_qp_in, in, qpc_data_ext);
+
+	DEVX_SET(init2init_qp_in, in, opcode, MLX5_CMD_OP_INIT2INIT_QP);
+	DEVX_SET(init2init_qp_in, in, qpc_ext, 1);
+	DEVX_SET(init2init_qp_in, in, qpn, qp->qp_num);
+	DEVX_SET64(init2init_qp_in, in, opt_param_mask_95_32, mask);
+
+	DEVX_SET(qpc_ext, qpce, qos_queue_group_id_requester, req_id);
+	DEVX_SET(qpc_ext, qpce, qos_queue_group_id_responder, resp_id);
+
+	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
+}
+
+static int modify_ib_qp_sched_elem_rts(struct ibv_qp *qp,
+				       uint32_t req_id, uint32_t resp_id)
+{
+	uint64_t mask = MLX5_QPC_OPT_MASK_32_QOS_QUEUE_GROUP_ID;
+	uint32_t in[DEVX_ST_SZ_DW(rts2rts_qp_in)] = {};
+	uint32_t out[DEVX_ST_SZ_DW(rts2rts_qp_out)] = {};
+	void *qpce = DEVX_ADDR_OF(rts2rts_qp_in, in, qpc_data_ext);
+
+	DEVX_SET(rts2rts_qp_in, in, opcode, MLX5_CMD_OP_RTS2RTS_QP);
+	DEVX_SET(rts2rts_qp_in, in, qpc_ext, 1);
+	DEVX_SET(rts2rts_qp_in, in, qpn, qp->qp_num);
+	DEVX_SET64(rts2rts_qp_in, in, opt_param_mask_95_32, mask);
+
+	DEVX_SET(qpc_ext, qpce, qos_queue_group_id_requester, req_id);
+	DEVX_SET(qpc_ext, qpce, qos_queue_group_id_responder, resp_id);
+
+	return mlx5dv_devx_qp_modify(qp, in, sizeof(in), out, sizeof(out));
+}
+
+static int modify_ib_qp_sched_elem(struct ibv_qp *qp,
+				   uint32_t req_id, uint32_t resp_id)
+{
+	int ret;
+
+	switch (qp->state) {
+	case IBV_QPS_INIT:
+		ret = modify_ib_qp_sched_elem_init(qp, req_id, resp_id);
+		break;
+
+	case IBV_QPS_RTS:
+		ret = modify_ib_qp_sched_elem_rts(qp, req_id, resp_id);
+		break;
+
+	default:
+		return EOPNOTSUPP;
+	};
+
+	return ret;
+}
+
+static int modify_raw_qp_sched_elem(struct ibv_qp *qp, uint32_t qos_id)
+{
+	struct mlx5_qos_caps *qc = &to_mctx(qp->context)->qos_caps;
+	uint32_t mout[DEVX_ST_SZ_DW(modify_sq_out)] = {};
+	uint32_t min[DEVX_ST_SZ_DW(modify_sq_in)] = {};
+	struct mlx5_qp *mqp = to_mqp(qp);
+	void *sqc;
+
+	if (qp->state != IBV_QPS_RTS || !qc->nic_sq_scheduling)
+		return EOPNOTSUPP;
+
+	DEVX_SET(modify_sq_in, min, opcode, MLX5_CMD_OP_MODIFY_SQ);
+	DEVX_SET(modify_sq_in, min, sq_state, MLX5_SQC_STATE_RDY);
+	DEVX_SET(modify_sq_in, min, sqn, mqp->sqn);
+	DEVX_SET64(modify_sq_in, min, modify_bitmask,
+		   MLX5_MODIFY_SQ_BITMASK_QOS_QUEUE_GROUP_ID);
+	sqc = DEVX_ADDR_OF(modify_sq_in, min, sq_context);
+	DEVX_SET(sqc, sqc, state, MLX5_SQC_STATE_RDY);
+	DEVX_SET(sqc, sqc, qos_queue_group_id, qos_id);
+
+	return mlx5dv_devx_qp_modify(qp, min, sizeof(min), mout, sizeof(mout));
+}
+
+int mlx5dv_modify_qp_sched_elem(struct ibv_qp *qp,
+				const struct mlx5dv_sched_leaf *requestor,
+				const struct mlx5dv_sched_leaf *responder)
+{
+	struct mlx5_qos_caps *qc = &to_mctx(qp->context)->qos_caps;
+
+	switch (qp->qp_type) {
+	case IBV_QPT_UC:
+	case IBV_QPT_UD:
+		if (responder)
+			return EINVAL;
+		SWITCH_FALLTHROUGH;
+	case IBV_QPT_RC:
+		if ((!to_mctx(qp->context)->qpc_extension_cap) ||
+		    !(qc->nic_qp_scheduling))
+			return EOPNOTSUPP;
+		return modify_ib_qp_sched_elem(qp,
+					       requestor ? requestor->obj->object_id : 0,
+					       responder ? responder->obj->object_id : 0);
+	case IBV_QPT_RAW_PACKET:
+		if (responder)
+			return EINVAL;
+		return modify_raw_qp_sched_elem(qp,
+						requestor ? requestor->obj->object_id : 0);
+	default:
+		return EOPNOTSUPP;
+	}
+}
+
 LATEST_SYMVER_FUNC(mlx5dv_init_obj, 1_2, "MLX5_1.2",
 		   int,
 		   struct mlx5dv_obj *obj, uint64_t obj_type)
@@ -1004,12 +1592,12 @@ COMPAT_SYMVER_FUNC(mlx5dv_init_obj, 1_0, "MLX5_1.0",
 		/* ABI version 1.0 returns the void ** in this memory
 		 * location
 		 */
-		obj->cq.out->cq_uar = to_mctx(obj->cq.in->context)->uar;
+		obj->cq.out->cq_uar = &(to_mctx(obj->cq.in->context)->cq_uar_reg);
 	}
 	return ret;
 }
 
-static off_t get_uar_mmap_offset(int idx, int page_size, int command)
+off_t get_uar_mmap_offset(int idx, int page_size, int command)
 {
 	off_t offset = 0;
 
@@ -1117,17 +1705,17 @@ repeat:
 
 static void adjust_uar_info(struct mlx5_device *mdev,
 			    struct mlx5_context *context,
-			    struct mlx5_alloc_ucontext_resp resp)
+			    struct mlx5_ib_alloc_ucontext_resp *resp)
 {
-	if (!resp.log_uar_size && !resp.num_uars_per_page) {
+	if (!resp->log_uar_size && !resp->num_uars_per_page) {
 		/* old kernel */
 		context->uar_size = mdev->page_size;
 		context->num_uars_per_page = 1;
 		return;
 	}
 
-	context->uar_size = 1 << resp.log_uar_size;
-	context->num_uars_per_page = resp.num_uars_per_page;
+	context->uar_size = 1 << resp->log_uar_size;
+	context->num_uars_per_page = resp->num_uars_per_page;
 }
 
 bool mlx5dv_is_supported(struct ibv_device *device)
@@ -1146,104 +1734,103 @@ mlx5dv_open_device(struct ibv_device *device, struct mlx5dv_context_attr *attr)
 	return verbs_open_device(device, attr);
 }
 
-static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
-						int cmd_fd,
-						void *private_data)
+static int get_uar_info(struct mlx5_device *mdev,
+			int *tot_uuars, int *low_lat_uuars)
 {
-	struct mlx5_context	       *context;
-	struct mlx5_alloc_ucontext	req;
-	struct mlx5_alloc_ucontext_resp resp;
-	int				i;
-	int				page_size;
-	int				tot_uuars;
-	int				low_lat_uuars;
-	int				gross_uuars;
-	int				j;
-	struct mlx5_device	       *mdev = to_mdev(ibdev);
-	struct verbs_context	       *v_ctx;
-	struct ibv_port_attr		port_attr;
-	struct ibv_device_attr_ex	device_attr;
-	int				k;
-	int				bfi;
-	int				num_sys_page_map;
-	struct mlx5dv_context_attr      *ctx_attr = private_data;
-
-	if (ctx_attr && ctx_attr->comp_mask) {
-		errno = EINVAL;
-		return NULL;
+	*tot_uuars = get_total_uuars(mdev->page_size);
+	if (*tot_uuars < 0) {
+		errno = -*tot_uuars;
+		return -1;
 	}
 
+	*low_lat_uuars = get_num_low_lat_uuars(*tot_uuars);
+	if (*low_lat_uuars < 0) {
+		errno = -*low_lat_uuars;
+		return -1;
+	}
+
+	if (*low_lat_uuars > *tot_uuars - 1) {
+		errno = ENOMEM;
+		return -1;
+	}
+
+	return 0;
+}
+
+static void mlx5_uninit_context(struct mlx5_context *context)
+{
+	close_debug_file(context);
+
+	verbs_uninit_context(&context->ibv_ctx);
+	free(context);
+}
+
+static struct mlx5_context *mlx5_init_context(struct ibv_device *ibdev,
+						int cmd_fd)
+{
+	struct mlx5_device *mdev = to_mdev(ibdev);
+	struct mlx5_context *context;
+	int low_lat_uuars;
+	int tot_uuars;
+	int ret;
+
 	context = verbs_init_and_alloc_context(ibdev, cmd_fd, context, ibv_ctx,
 					       RDMA_DRIVER_MLX5);
 	if (!context)
 		return NULL;
 
-	v_ctx = &context->ibv_ctx;
-	page_size = mdev->page_size;
-	mlx5_single_threaded = single_threaded_app();
-
 	open_debug_file(context);
 	set_debug_mask();
 	set_freeze_on_error();
 	if (gethostname(context->hostname, sizeof(context->hostname)))
 		strcpy(context->hostname, "host_unknown");
 
-	tot_uuars = get_total_uuars(page_size);
-	if (tot_uuars < 0) {
-		errno = -tot_uuars;
-		goto err_free;
+	mlx5_single_threaded = single_threaded_app();
+
+	ret = get_uar_info(mdev, &tot_uuars, &low_lat_uuars);
+	if (ret) {
+		mlx5_uninit_context(context);
+		return NULL;
 	}
+	context->tot_uuars = tot_uuars;
+	context->low_lat_uuars = low_lat_uuars;
 
-	low_lat_uuars = get_num_low_lat_uuars(tot_uuars);
-	if (low_lat_uuars < 0) {
-		errno = -low_lat_uuars;
-		goto err_free;
-	}
+	return context;
+}
 
-	if (low_lat_uuars > tot_uuars - 1) {
-		errno = ENOMEM;
-		goto err_free;
-	}
+static int mlx5_set_context(struct mlx5_context *context,
+			    struct mlx5_ib_alloc_ucontext_resp *resp,
+			    bool is_import)
+{
+	struct verbs_context *v_ctx = &context->ibv_ctx;
+	struct ibv_port_attr port_attr = {};
+	int cmd_fd = v_ctx->context.cmd_fd;
+	struct mlx5_device *mdev = to_mdev(v_ctx->context.device);
+	struct ibv_device *ibdev = v_ctx->context.device;
+	int page_size = mdev->page_size;
+	int num_sys_page_map;
+	int gross_uuars;
+	int bfi;
+	int i, k, j;
 
-	memset(&req, 0, sizeof(req));
-	memset(&resp, 0, sizeof(resp));
+	context->max_num_qps = resp->qp_tab_size;
+	context->bf_reg_size = resp->bf_reg_size;
+	context->cache_line_size = resp->cache_line_size;
+	context->max_sq_desc_sz = resp->max_sq_desc_sz;
+	context->max_rq_desc_sz = resp->max_rq_desc_sz;
+	context->max_send_wqebb	= resp->max_send_wqebb;
+	context->num_ports = resp->num_ports;
+	context->max_recv_wr = resp->max_recv_wr;
+	context->max_srq_recv_wr = resp->max_srq_recv_wr;
+	context->num_dyn_bfregs = resp->num_dyn_bfregs;
 
-	req.total_num_bfregs = tot_uuars;
-	req.num_low_latency_bfregs = low_lat_uuars;
-	req.max_cqe_version = MLX5_CQE_VERSION_V1;
-	req.lib_caps |= MLX5_LIB_CAP_4K_UAR;
-	if (ctx_attr && ctx_attr->flags) {
+	if (resp->comp_mask & MLX5_IB_ALLOC_UCONTEXT_RESP_MASK_ECE)
+		context->flags |= MLX5_CTX_FLAGS_ECE_SUPPORTED;
 
-		if (!check_comp_mask(ctx_attr->flags,
-				     MLX5DV_CONTEXT_FLAGS_DEVX)) {
-			errno = EINVAL;
-			goto err_free;
-		}
-
-		req.flags = MLX5_IB_ALLOC_UCTX_DEVX;
-	}
-
-	if (mlx5_cmd_get_context(context, &req, sizeof(req), &resp,
-				 sizeof(resp)))
-		goto err_free;
-
-	context->max_num_qps		= resp.qp_tab_size;
-	context->bf_reg_size		= resp.bf_reg_size;
-	context->tot_uuars		= resp.tot_bfregs;
-	context->low_lat_uuars		= low_lat_uuars;
-	context->cache_line_size	= resp.cache_line_size;
-	context->max_sq_desc_sz = resp.max_sq_desc_sz;
-	context->max_rq_desc_sz = resp.max_rq_desc_sz;
-	context->max_send_wqebb	= resp.max_send_wqebb;
-	context->num_ports	= resp.num_ports;
-	context->max_recv_wr	= resp.max_recv_wr;
-	context->max_srq_recv_wr = resp.max_srq_recv_wr;
-	context->num_dyn_bfregs = resp.num_dyn_bfregs;
-
-	if (resp.comp_mask & MLX5_IB_ALLOC_UCONTEXT_RESP_MASK_DUMP_FILL_MKEY) {
-		context->dump_fill_mkey = resp.dump_fill_mkey;
+	if (resp->comp_mask & MLX5_IB_ALLOC_UCONTEXT_RESP_MASK_DUMP_FILL_MKEY) {
+		context->dump_fill_mkey = resp->dump_fill_mkey;
 		/* Have the BE value ready to be used in data path */
-		context->dump_fill_mkey_be = htobe32(resp.dump_fill_mkey);
+		context->dump_fill_mkey_be = htobe32(resp->dump_fill_mkey);
 	} else {
 		/* kernel driver will never return MLX5_INVALID_LKEY for
 		 * dump_fill_mkey
@@ -1252,33 +1839,17 @@ static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
 		context->dump_fill_mkey_be = htobe32(MLX5_INVALID_LKEY);
 	}
 
-	if (context->num_dyn_bfregs) {
-		context->count_dyn_bfregs = calloc(context->num_dyn_bfregs,
-						   sizeof(*context->count_dyn_bfregs));
-		if (!context->count_dyn_bfregs) {
-			errno = ENOMEM;
-			goto err_free;
-		}
-	}
-
-	context->cqe_version = resp.cqe_version;
-
+	context->cqe_version = resp->cqe_version;
 	adjust_uar_info(mdev, context, resp);
 
-	gross_uuars = context->tot_uuars / MLX5_NUM_NON_FP_BFREGS_PER_UAR * NUM_BFREGS_PER_UAR;
-	context->bfs = calloc(gross_uuars + context->num_dyn_bfregs, sizeof(*context->bfs));
-
-	if (!context->bfs) {
-		errno = ENOMEM;
-		goto err_free;
-	}
-
-	context->cmds_supp_uhw = resp.cmds_supp_uhw;
+	context->cmds_supp_uhw = resp->cmds_supp_uhw;
 	context->vendor_cap_flags = 0;
-	context->start_dyn_bfregs_index = gross_uuars;
+	list_head_init(&context->dyn_uar_bf_list);
+	list_head_init(&context->dyn_uar_qp_shared_list);
+	list_head_init(&context->dyn_uar_qp_dedicated_list);
 
-	if (resp.eth_min_inline)
-		context->eth_min_inline_size = (resp.eth_min_inline == MLX5_USER_INLINE_MODE_NONE) ?
+	if (resp->eth_min_inline)
+		context->eth_min_inline_size = (resp->eth_min_inline == MLX5_USER_INLINE_MODE_NONE) ?
 						0 : MLX5_ETH_L2_INLINE_HEADER_SIZE;
 	else
 		context->eth_min_inline_size = MLX5_ETH_L2_INLINE_HEADER_SIZE;
@@ -1300,6 +1871,27 @@ static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
 	context->prefer_bf = get_always_bf();
 	context->shut_up_bf = get_shut_up_bf();
 
+	if (resp->tot_bfregs) {
+		if (is_import) {
+			errno = EINVAL;
+			return EINVAL;
+		}
+		context->tot_uuars = resp->tot_bfregs;
+		gross_uuars = context->tot_uuars / MLX5_NUM_NON_FP_BFREGS_PER_UAR * NUM_BFREGS_PER_UAR;
+		context->bfs = calloc(gross_uuars, sizeof(*context->bfs));
+		if (!context->bfs) {
+			errno = ENOMEM;
+			goto err_free;
+		}
+		context->flags |= MLX5_CTX_FLAGS_NO_KERN_DYN_UAR;
+	} else {
+		context->qp_max_dedicated_uuars = context->low_lat_uuars;
+		context->qp_max_shared_uuars = context->tot_uuars - context->low_lat_uuars;
+		goto bf_done;
+	}
+
+	context->max_num_legacy_dyn_uar_sys_page = context->num_dyn_bfregs /
+			(context->num_uars_per_page * MLX5_NUM_NON_FP_BFREGS_PER_UAR);
 	num_sys_page_map = context->tot_uuars / (context->num_uars_per_page * MLX5_NUM_NON_FP_BFREGS_PER_UAR);
 	for (i = 0; i < num_sys_page_map; ++i) {
 		if (mlx5_mmap(&context->uar[i], i, cmd_fd, page_size,
@@ -1322,30 +1914,26 @@ static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
 				if (bfi)
 					context->bfs[bfi].buf_size = context->bf_reg_size / 2;
 				context->bfs[bfi].uuarn = bfi;
-				context->bfs[bfi].uar_mmap_offset = get_uar_mmap_offset(i,
-											page_size,
-											uar_type_to_cmd(context->uar[i].type));
+				context->bfs[bfi].uar_mmap_offset =
+					get_uar_mmap_offset(i, page_size,
+							uar_type_to_cmd(context->uar[i].type));
 			}
 		}
 	}
+
+bf_done:
+
 	context->hca_core_clock = NULL;
-	if (resp.response_length + sizeof(resp.ibv_resp) >=
-	    offsetof(struct mlx5_alloc_ucontext_resp, hca_core_clock_offset) +
-	    sizeof(resp.hca_core_clock_offset) &&
-	    resp.comp_mask & MLX5_IB_ALLOC_UCONTEXT_RESP_MASK_CORE_CLOCK_OFFSET) {
-		context->core_clock.offset = resp.hca_core_clock_offset;
+	if (resp->comp_mask & MLX5_IB_ALLOC_UCONTEXT_RESP_MASK_CORE_CLOCK_OFFSET) {
+		context->core_clock.offset = resp->hca_core_clock_offset;
 		mlx5_map_internal_clock(mdev, &v_ctx->context);
 	}
 
 	context->clock_info_page = NULL;
-	if (resp.response_length + sizeof(resp.ibv_resp) >=
-	    offsetof(struct mlx5_alloc_ucontext_resp, clock_info_versions) +
-	    sizeof(resp.clock_info_versions) &&
-	    (resp.clock_info_versions & (1 << MLX5_IB_CLOCK_INFO_V1))) {
+	if ((resp->clock_info_versions & (1 << MLX5_IB_CLOCK_INFO_V1)))
 		mlx5_map_clock_info(mdev, &v_ctx->context);
-	}
 
-	context->flow_action_flags = resp.flow_action_flags;
+	context->flow_action_flags = resp->flow_action_flags;
 
 	mlx5_read_env(ibdev, context);
 
@@ -1360,15 +1948,7 @@ static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
 			goto err_free;
 	}
 
-	memset(&device_attr, 0, sizeof(device_attr));
-	if (!mlx5_query_device_ex(&v_ctx->context, NULL, &device_attr,
-				  sizeof(struct ibv_device_attr_ex))) {
-		context->cached_device_cap_flags =
-			device_attr.orig_attr.device_cap_flags;
-		context->atomic_cap = device_attr.orig_attr.atomic_cap;
-		context->cached_tso_caps = device_attr.tso_caps;
-		context->max_dm_size = device_attr.max_dm_size;
-	}
+	mlx5_query_device_ctx(context);
 
 	for (j = 0; j < min(MLX5_MAX_PORTS_NUM, context->num_ports); ++j) {
 		memset(&port_attr, 0, sizeof(port_attr));
@@ -1378,21 +1958,116 @@ static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
 		}
 	}
 
-	return v_ctx;
+	mlx5_set_singleton_nc_uar(&v_ctx->context);
+	context->cq_uar_reg = context->nc_uar ? context->nc_uar->uar : context->uar[0].reg;
+
+	return 0;
 
 err_free_bf:
 	free(context->bfs);
 
 err_free:
-	free(context->count_dyn_bfregs);
 	for (i = 0; i < MLX5_MAX_UARS; ++i) {
 		if (context->uar[i].reg)
 			munmap(context->uar[i].reg, page_size);
 	}
-	close_debug_file(context);
 
-	verbs_uninit_context(&context->ibv_ctx);
-	free(context);
+	return -1;
+}
+
+static struct verbs_context *mlx5_alloc_context(struct ibv_device *ibdev,
+						int cmd_fd,
+						void *private_data)
+{
+	struct mlx5_context	       *context;
+	struct mlx5_alloc_ucontext	req = {};
+	struct mlx5_alloc_ucontext_resp resp = {};
+	struct mlx5dv_context_attr      *ctx_attr = private_data;
+	bool				always_devx = false;
+	int ret;
+
+	context = mlx5_init_context(ibdev, cmd_fd);
+	if (!context)
+		return NULL;
+
+	if (ctx_attr && ctx_attr->comp_mask) {
+		errno = EINVAL;
+		goto err;
+	}
+
+	req.total_num_bfregs = context->tot_uuars;
+	req.num_low_latency_bfregs = context->low_lat_uuars;
+	req.max_cqe_version = MLX5_CQE_VERSION_V1;
+	req.lib_caps |= (MLX5_LIB_CAP_4K_UAR | MLX5_LIB_CAP_DYN_UAR);
+	if (ctx_attr && ctx_attr->flags) {
+
+		if (!check_comp_mask(ctx_attr->flags,
+				     MLX5DV_CONTEXT_FLAGS_DEVX)) {
+			errno = EINVAL;
+			goto err;
+		}
+
+		req.flags = MLX5_IB_ALLOC_UCTX_DEVX;
+	} else {
+		req.flags = MLX5_IB_ALLOC_UCTX_DEVX;
+		always_devx = true;
+	}
+
+retry_open:
+	if (mlx5_cmd_get_context(context, &req, sizeof(req), &resp,
+				 sizeof(resp))) {
+		if (always_devx) {
+			req.flags &= ~MLX5_IB_ALLOC_UCTX_DEVX;
+			always_devx = false;
+			memset(&resp, 0, sizeof(resp));
+			goto retry_open;
+		} else {
+			goto err;
+		}
+	}
+
+	ret = mlx5_set_context(context, &resp.drv_payload, false);
+	if (ret)
+		goto err;
+
+	return &context->ibv_ctx;
+
+err:
+	mlx5_uninit_context(context);
+	return NULL;
+}
+
+static struct verbs_context *mlx5_import_context(struct ibv_device *ibdev,
+						int cmd_fd)
+
+{
+	struct mlx5_ib_alloc_ucontext_resp resp = {};
+	DECLARE_COMMAND_BUFFER_LINK(driver_attr, UVERBS_OBJECT_DEVICE,
+				    UVERBS_METHOD_QUERY_CONTEXT, 1,
+				    NULL);
+	struct ibv_context *context;
+	struct mlx5_context *mctx;
+	int ret;
+
+	mctx = mlx5_init_context(ibdev, cmd_fd);
+	if (!mctx)
+		return NULL;
+
+	context = &mctx->ibv_ctx.context;
+
+	fill_attr_out_ptr(driver_attr, MLX5_IB_ATTR_QUERY_CONTEXT_RESP_UCTX, &resp);
+	ret = ibv_cmd_query_context(context, driver_attr);
+	if (ret)
+		goto err;
+
+	ret = mlx5_set_context(mctx, &resp, true);
+	if (ret)
+		goto err;
+
+	return &mctx->ibv_ctx;
+
+err:
+	mlx5_uninit_context(mctx);
 	return NULL;
 }
 
@@ -1402,13 +2077,6 @@ static void mlx5_free_context(struct ibv_context *ibctx)
 	int page_size = to_mdev(ibctx->device)->page_size;
 	int i;
 
-	for (i = context->start_dyn_bfregs_index;
-	      i < context->start_dyn_bfregs_index + context->num_dyn_bfregs; i++) {
-		if (context->bfs[i].uar)
-			munmap(context->bfs[i].uar, page_size);
-	}
-
-	free(context->count_dyn_bfregs);
 	free(context->bfs);
 	for (i = 0; i < MLX5_MAX_UARS; ++i) {
 		if (context->uar[i].reg)
@@ -1420,6 +2088,7 @@ static void mlx5_free_context(struct ibv_context *ibctx)
 	if (context->clock_info_page)
 		munmap((void *)context->clock_info_page, page_size);
 	close_debug_file(context);
+	clean_dyn_uars(ibctx);
 
 	verbs_uninit_context(&context->ibv_ctx);
 	free(context);
@@ -1454,6 +2123,7 @@ static const struct verbs_device_ops mlx5_dev_ops = {
 	.alloc_device = mlx5_device_alloc,
 	.uninit_device = mlx5_uninit_device,
 	.alloc_context = mlx5_alloc_context,
+	.import_context = mlx5_import_context,
 };
 
 bool is_mlx5_dev(struct ibv_device *device)
