@@ -628,13 +628,14 @@ struct mlx5_ifc_cmd_hca_cap_bits {
 
 	u8         log_max_srq_sz[0x8];
 	u8         log_max_qp_sz[0x8];
-	u8         reserved_at_90[0x8];
+	u8         reserved_at_90[0x3];
+	u8         isolate_vl_tc_new[0x1];
+	u8         reserved_at_94[0x4];
 	u8         prio_tag_required[0x1];
 	u8         reserved_at_99[0x2];
 	u8         log_max_qp[0x5];
 
-	u8         reserved_at_a0[0xa];
-	u8         isolate_vl_tc[0x1];
+	u8         reserved_at_a0[0xb];
 	u8         log_max_srq[0x5];
 	u8         reserved_at_b0[0x10];
 
