@@ -41,11 +41,16 @@
 #include "mlx5_ifc.h"
 #include "mlx5.h"
 
-#define DR_RULE_MAX_STES	17
-#define DR_ACTION_MAX_STES	3
+#define DR_RULE_MAX_STES	20
+#define DR_ACTION_MAX_STES	6
 #define WIRE_PORT		0xFFFF
 #define DR_STE_SVLAN		0x1
 #define DR_STE_CVLAN		0x2
+#define CVLAN_ETHERTYPE	0x8100
+#define SVLAN_ETHERTYPE	0x88a8
+#define NUM_OF_FLEX_PARSERS	8
+#define DR_STE_MAX_FLEX_0_ID	3
+#define DR_STE_MAX_FLEX_1_ID	7
 
 #define dr_dbg(dmn, arg...) dr_dbg_ctx((dmn)->ctx, ##arg)
 
@@ -95,47 +100,7 @@ dr_icm_next_higher_chunk(enum dr_icm_chunk_size chunk)
 }
 
 enum dr_ste_lu_type {
-	DR_STE_LU_TYPE_NOP			= 0x00,
-	DR_STE_LU_TYPE_SRC_GVMI_AND_QP		= 0x05,
-	DR_STE_LU_TYPE_ETHL2_TUNNELING_I	= 0x0a,
-	DR_STE_LU_TYPE_ETHL2_DST_O		= 0x06,
-	DR_STE_LU_TYPE_ETHL2_DST_I		= 0x07,
-	DR_STE_LU_TYPE_ETHL2_DST_D		= 0x1b,
-	DR_STE_LU_TYPE_ETHL2_SRC_O		= 0x08,
-	DR_STE_LU_TYPE_ETHL2_SRC_I		= 0x09,
-	DR_STE_LU_TYPE_ETHL2_SRC_D		= 0x1c,
-	DR_STE_LU_TYPE_ETHL2_SRC_DST_O		= 0x36,
-	DR_STE_LU_TYPE_ETHL2_SRC_DST_I		= 0x37,
-	DR_STE_LU_TYPE_ETHL2_SRC_DST_D		= 0x38,
-	DR_STE_LU_TYPE_ETHL3_IPV6_DST_O		= 0x0d,
-	DR_STE_LU_TYPE_ETHL3_IPV6_DST_I		= 0x0e,
-	DR_STE_LU_TYPE_ETHL3_IPV6_DST_D		= 0x1e,
-	DR_STE_LU_TYPE_ETHL3_IPV6_SRC_O		= 0x0f,
-	DR_STE_LU_TYPE_ETHL3_IPV6_SRC_I		= 0x10,
-	DR_STE_LU_TYPE_ETHL3_IPV6_SRC_D		= 0x1f,
-	DR_STE_LU_TYPE_ETHL3_IPV4_5_TUPLE_O	= 0x11,
-	DR_STE_LU_TYPE_ETHL3_IPV4_5_TUPLE_I	= 0x12,
-	DR_STE_LU_TYPE_ETHL3_IPV4_5_TUPLE_D	= 0x20,
-	DR_STE_LU_TYPE_ETHL3_IPV4_MISC_O	= 0x29,
-	DR_STE_LU_TYPE_ETHL3_IPV4_MISC_I	= 0x2a,
-	DR_STE_LU_TYPE_ETHL3_IPV4_MISC_D	= 0x2b,
-	DR_STE_LU_TYPE_ETHL4_O			= 0x13,
-	DR_STE_LU_TYPE_ETHL4_I			= 0x14,
-	DR_STE_LU_TYPE_ETHL4_D			= 0x21,
-	DR_STE_LU_TYPE_ETHL4_MISC_O		= 0x2c,
-	DR_STE_LU_TYPE_ETHL4_MISC_I		= 0x2d,
-	DR_STE_LU_TYPE_ETHL4_MISC_D		= 0x2e,
-	DR_STE_LU_TYPE_MPLS_FIRST_O		= 0x15,
-	DR_STE_LU_TYPE_MPLS_FIRST_I		= 0x24,
-	DR_STE_LU_TYPE_MPLS_FIRST_D		= 0x25,
-	DR_STE_LU_TYPE_GRE			= 0x16,
-	DR_STE_LU_TYPE_FLEX_PARSER_0		= 0x22,
-	DR_STE_LU_TYPE_FLEX_PARSER_1		= 0x23,
-	DR_STE_LU_TYPE_FLEX_PARSER_TNL_HEADER	= 0x19,
-	DR_STE_LU_TYPE_GENERAL_PURPOSE		= 0x18,
-	DR_STE_LU_TYPE_STEERING_REGISTERS_0	= 0x2f,
-	DR_STE_LU_TYPE_STEERING_REGISTERS_1	= 0x30,
-	DR_STE_LU_TYPE_DONT_CARE		= 0x0f,
+	DR_STE_LU_TYPE_DONT_CARE	= 0x0f,
 };
 
 enum dr_ste_entry_type {
@@ -149,6 +114,7 @@ enum {
 	DR_STE_SIZE_CTRL	= 32,
 	DR_STE_SIZE_TAG		= 16,
 	DR_STE_SIZE_MASK	= 16,
+	DR_STE_LOG_SIZE		= 6,
 };
 
 enum {
@@ -157,6 +123,7 @@ enum {
 
 enum {
 	DR_MODIFY_ACTION_SIZE	= 8,
+	DR_MODIFY_ACTION_LOG_SIZE	= 3,
 };
 
 enum dr_matcher_criteria {
@@ -166,7 +133,8 @@ enum dr_matcher_criteria {
 	DR_MATCHER_CRITERIA_INNER	= 1 << 2,
 	DR_MATCHER_CRITERIA_MISC2	= 1 << 3,
 	DR_MATCHER_CRITERIA_MISC3	= 1 << 4,
-	DR_MATCHER_CRITERIA_MAX		= 1 << 5,
+	DR_MATCHER_CRITERIA_MISC4	= 1 << 5,
+	DR_MATCHER_CRITERIA_MAX		= 1 << 6,
 };
 
 enum dr_action_type {
@@ -183,6 +151,12 @@ enum dr_action_type {
 	DR_ACTION_TYP_VPORT,
 	DR_ACTION_TYP_METER,
 	DR_ACTION_TYP_MISS,
+	DR_ACTION_TYP_SAMPLER,
+	DR_ACTION_TYP_DEST_ARRAY,
+	DR_ACTION_TYP_POP_VLAN,
+	DR_ACTION_TYP_PUSH_VLAN,
+	DR_ACTION_TYP_ASO_FIRST_HIT,
+	DR_ACTION_TYP_ASO_FLOW_METER,
 	DR_ACTION_TYP_MAX,
 };
 
@@ -192,7 +166,9 @@ struct dr_icm_buddy_mem;
 struct dr_ste_htbl;
 struct dr_match_param;
 struct dr_devx_caps;
+struct dr_rule_rx_tx;
 struct dr_matcher_rx_tx;
+struct dr_ste_ctx;
 
 struct dr_data_seg {
 	uint64_t	addr;
@@ -216,14 +192,14 @@ struct dr_ste {
 	/* attached to the miss_list head at each htbl entry */
 	struct list_node	miss_list_node;
 
-	/* each rule member that uses this ste attached here */
-	struct list_head	rule_list;
-
 	/* this ste is member of htbl */
 	struct dr_ste_htbl	*htbl;
 
 	struct dr_ste_htbl	*next_htbl;
 
+	/* The rule this STE belongs to */
+	struct dr_rule_rx_tx    *rule_rx_tx;
+
 	/* this ste is part of a rule, located in ste's chain */
 	uint8_t			ste_chain_location;
 };
@@ -241,7 +217,7 @@ struct dr_ste_htbl_ctrl {
 };
 
 struct dr_ste_htbl {
-	uint8_t			lu_type;
+	uint16_t		lu_type;
 	uint16_t		byte_mask;
 	atomic_int		refcount;
 	struct dr_icm_chunk	*chunk;
@@ -275,7 +251,7 @@ struct dr_ste_build {
 	bool			inner;
 	bool			rx;
 	struct dr_devx_caps	*caps;
-	uint8_t			lu_type;
+	uint16_t		lu_type;
 	uint16_t		byte_mask;
 	uint8_t			bit_mask[DR_STE_SIZE_MASK];
 	int (*ste_build_tag_func)(struct dr_match_param *spec,
@@ -285,7 +261,7 @@ struct dr_ste_build {
 
 struct dr_ste_htbl *dr_ste_htbl_alloc(struct dr_icm_pool *pool,
 				      enum dr_icm_chunk_size chunk_size,
-				      uint8_t lu_type, uint16_t byte_mask);
+				      uint16_t lu_type, uint16_t byte_mask);
 int dr_ste_htbl_free(struct dr_ste_htbl *htbl);
 
 static inline void dr_htbl_put(struct dr_ste_htbl *htbl)
@@ -301,28 +277,94 @@ static inline void dr_htbl_get(struct dr_ste_htbl *htbl)
 
 /* STE utils */
 uint32_t dr_ste_calc_hash_index(uint8_t *hw_ste_p, struct dr_ste_htbl *htbl);
-void dr_ste_init(uint8_t *hw_ste_p, uint8_t lu_type, uint8_t entry_type, uint16_t gvmi);
-void dr_ste_always_hit_htbl(struct dr_ste *ste, struct dr_ste_htbl *next_htbl);
-void dr_ste_set_miss_addr(uint8_t *hw_ste, uint64_t miss_addr);
-uint64_t dr_ste_get_miss_addr(uint8_t *hw_ste);
-void dr_ste_set_hit_addr(uint8_t *hw_ste, uint64_t icm_addr, uint32_t ht_size);
-void dr_ste_always_miss_addr(struct dr_ste *ste, uint64_t miss_addr);
+void dr_ste_set_miss_addr(struct dr_ste_ctx *ste_ctx, uint8_t *hw_ste_p,
+			  uint64_t miss_addr);
+void dr_ste_set_hit_addr_by_next_htbl(struct dr_ste_ctx *ste_ctx,
+				      uint8_t *hw_ste,
+				      struct dr_ste_htbl *next_htbl);
+void dr_ste_set_hit_addr(struct dr_ste_ctx *ste_ctx, uint8_t *hw_ste_p,
+			 uint64_t icm_addr, uint32_t ht_size);
 void dr_ste_set_bit_mask(uint8_t *hw_ste_p, uint8_t *bit_mask);
 bool dr_ste_is_last_in_rule(struct dr_matcher_rx_tx *nic_matcher,
 			    uint8_t ste_location);
-void dr_ste_rx_set_flow_tag(uint8_t *hw_ste_p, uint32_t flow_tag);
-void dr_ste_set_counter_id(uint8_t *hw_ste_p, uint32_t ctr_id);
-void dr_ste_set_tx_encap(void *hw_ste_p, uint32_t reformat_id, int size, bool encap_l3);
-void dr_ste_set_rx_decap(uint8_t *hw_ste_p);
-void dr_ste_set_rx_decap_l3(uint8_t *hw_ste_p, bool vlan);
-void dr_ste_set_entry_type(uint8_t *hw_ste_p, uint8_t entry_type);
-uint8_t dr_ste_get_entry_type(uint8_t *hw_ste_p);
-void dr_ste_set_rewrite_actions(uint8_t *hw_ste_p, uint16_t num_of_actions,
-				uint32_t re_write_index);
 uint64_t dr_ste_get_icm_addr(struct dr_ste *ste);
 uint64_t dr_ste_get_mr_addr(struct dr_ste *ste);
 struct list_head *dr_ste_get_miss_list(struct dr_ste *ste);
 
+#define MAX_VLANS 2
+
+struct dr_action_aso {
+	struct mlx5dv_devx_obj *devx_obj;
+	uint32_t offset;
+	uint8_t dest_reg_id;
+	union {
+		struct {
+			bool set;
+		} first_hit;
+		struct {
+			uint8_t initial_color;
+		} flow_meter;
+	};
+};
+
+struct dr_ste_actions_attr {
+	uint32_t	modify_index;
+	uint16_t	modify_actions;
+	uint32_t	decap_index;
+	uint16_t	decap_actions;
+	bool		decap_with_vlan;
+	uint64_t	final_icm_addr;
+	uint32_t	flow_tag;
+	uint32_t	ctr_id;
+	uint16_t	gvmi;
+	uint16_t	hit_gvmi;
+	uint32_t	reformat_id;
+	uint32_t	reformat_size;
+	bool		prio_tag_required;
+	struct {
+		int		count;
+		uint32_t	headers[MAX_VLANS];
+	} vlans;
+	struct dr_action_aso *aso;
+};
+
+void dr_ste_set_actions_rx(struct dr_ste_ctx *ste_ctx,
+			   uint8_t *action_type_set,
+			   uint8_t *last_ste,
+			   struct dr_ste_actions_attr *attr,
+			   uint32_t *added_stes);
+void dr_ste_set_actions_tx(struct dr_ste_ctx *ste_ctx,
+			   uint8_t *action_type_set,
+			   uint8_t *last_ste,
+			   struct dr_ste_actions_attr *attr,
+			   uint32_t *added_stes);
+void dr_ste_set_action_set(struct dr_ste_ctx *ste_ctx,
+			   __be64 *hw_action,
+			   uint8_t hw_field,
+			   uint8_t shifter,
+			   uint8_t length,
+			   uint32_t data);
+void dr_ste_set_action_add(struct dr_ste_ctx *ste_ctx,
+			   __be64 *hw_action,
+			   uint8_t hw_field,
+			   uint8_t shifter,
+			   uint8_t length,
+			   uint32_t data);
+void dr_ste_set_action_copy(struct dr_ste_ctx *ste_ctx,
+			    __be64 *hw_action,
+			    uint8_t dst_hw_field,
+			    uint8_t dst_shifter,
+			    uint8_t dst_len,
+			    uint8_t src_hw_field,
+			    uint8_t src_shifter);
+int dr_ste_set_action_decap_l3_list(struct dr_ste_ctx *ste_ctx,
+				   void *data, uint32_t data_sz,
+				   uint8_t *hw_action, uint32_t hw_action_sz,
+				   uint16_t *used_hw_action_num);
+const struct dr_ste_action_modify_field *
+dr_ste_conv_modify_hdr_sw_field(struct dr_ste_ctx *ste_ctx, uint16_t sw_field);
+
+struct dr_ste_ctx *dr_ste_get_ctx(uint8_t version);
 void dr_ste_free(struct dr_ste *ste,
 		 struct mlx5dv_dr_matcher *matcher,
 		 struct dr_matcher_rx_tx *nic_matcher);
@@ -345,8 +387,6 @@ static inline bool dr_ste_is_not_used(struct dr_ste *ste)
 	return !atomic_load(&ste->refcount);
 }
 
-void dr_ste_set_hit_addr_by_next_htbl(uint8_t *hw_ste,
-				      struct dr_ste_htbl *next_htbl);
 bool dr_ste_equal_tag(void *src, void *dst);
 int dr_ste_create_next_htbl(struct mlx5dv_dr_matcher *matcher,
 			    struct dr_matcher_rx_tx *nic_matcher,
@@ -363,71 +403,111 @@ int dr_ste_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			 struct dr_matcher_rx_tx *nic_matcher,
 			 struct dr_match_param *value,
 			 uint8_t *ste_arr);
-void dr_ste_build_eth_l2_src_dst(struct dr_ste_build *sb,
+void dr_ste_build_eth_l2_src_dst(struct dr_ste_ctx *ste_ctx,
+				 struct dr_ste_build *sb,
 				 struct dr_match_param *mask,
 				 bool inner, bool rx);
-void dr_ste_build_eth_l3_ipv4_5_tuple(struct dr_ste_build *sb,
+void dr_ste_build_eth_l3_ipv4_5_tuple(struct dr_ste_ctx *ste_ctx,
+				      struct dr_ste_build *sb,
 				      struct dr_match_param *mask,
 				      bool inner, bool rx);
-void dr_ste_build_eth_l3_ipv4_misc(struct dr_ste_build *sb,
+void dr_ste_build_eth_l3_ipv4_misc(struct dr_ste_ctx *ste_ctx,
+				   struct dr_ste_build *sb,
 				   struct dr_match_param *mask,
 				   bool inner, bool rx);
-void dr_ste_build_eth_l3_ipv6_dst(struct dr_ste_build *sb,
+void dr_ste_build_eth_l3_ipv6_dst(struct dr_ste_ctx *ste_ctx,
+				  struct dr_ste_build *sb,
 				  struct dr_match_param *mask,
 				  bool inner, bool rx);
-void dr_ste_build_eth_l3_ipv6_src(struct dr_ste_build *sb,
+void dr_ste_build_eth_l3_ipv6_src(struct dr_ste_ctx *ste_ctx,
+				  struct dr_ste_build *sb,
 				  struct dr_match_param *mask,
 				  bool inner, bool rx);
-void dr_ste_build_eth_l2_src(struct dr_ste_build *sb,
+void dr_ste_build_eth_l2_src(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_eth_l2_dst(struct dr_ste_build *sb,
+void dr_ste_build_eth_l2_dst(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_eth_l2_tnl(struct dr_ste_build *sb,
+void dr_ste_build_eth_l2_tnl(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_eth_ipv6_l3_l4(struct dr_ste_build *sb,
+void dr_ste_build_eth_ipv6_l3_l4(struct dr_ste_ctx *ste_ctx,
+				 struct dr_ste_build *sb,
 				 struct dr_match_param *mask,
 				 bool inner, bool rx);
-void dr_ste_build_eth_l4_misc(struct dr_ste_build *sb,
+void dr_ste_build_eth_l4_misc(struct dr_ste_ctx *ste_ctx,
+			      struct dr_ste_build *sb,
 			      struct dr_match_param *mask,
 			      bool inner, bool rx);
-void dr_ste_build_tnl_gre(struct dr_ste_build *sb,
+void dr_ste_build_tnl_gre(struct dr_ste_ctx *ste_ctx,
+			  struct dr_ste_build *sb,
 			  struct dr_match_param *mask,
 			  bool inner, bool rx);
-void dr_ste_build_mpls(struct dr_ste_build *sb,
+void dr_ste_build_mpls(struct dr_ste_ctx *ste_ctx,
+		       struct dr_ste_build *sb,
 		       struct dr_match_param *mask,
 		       bool inner, bool rx);
-void dr_ste_build_tnl_mpls(struct dr_ste_build *sb,
-			   struct dr_match_param *mask,
-			   bool inner, bool rx);
-int dr_ste_build_icmp(struct dr_ste_build *sb,
-		      struct dr_match_param *mask,
-		      struct dr_devx_caps *caps,
-		      bool inner, bool rx);
-void dr_ste_build_tnl_vxlan_gpe(struct dr_ste_build *sb,
+void dr_ste_build_tnl_mpls_over_gre(struct dr_ste_ctx *ste_ctx,
+				    struct dr_ste_build *sb,
+				    struct dr_match_param *mask,
+				    struct dr_devx_caps *caps,
+				    bool inner, bool rx);
+void dr_ste_build_tnl_mpls_over_udp(struct dr_ste_ctx *ste_ctx,
+				    struct dr_ste_build *sb,
+				    struct dr_match_param *mask,
+				    struct dr_devx_caps *caps,
+				    bool inner, bool rx);
+void dr_ste_build_icmp(struct dr_ste_ctx *ste_ctx,
+		       struct dr_ste_build *sb,
+		       struct dr_match_param *mask,
+		       struct dr_devx_caps *caps,
+		       bool inner, bool rx);
+void dr_ste_build_tnl_vxlan_gpe(struct dr_ste_ctx *ste_ctx,
+				struct dr_ste_build *sb,
 				struct dr_match_param *mask,
 				bool inner, bool rx);
-void dr_ste_build_tnl_geneve(struct dr_ste_build *sb,
+void dr_ste_build_tnl_geneve(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_tnl_gtpu(struct dr_ste_build *sb,
+void dr_ste_build_tnl_geneve_tlv_opt(struct dr_ste_ctx *ste_ctx,
+				     struct dr_ste_build *sb,
+				     struct dr_match_param *mask,
+				     struct dr_devx_caps *caps,
+				     bool inner, bool rx);
+void dr_ste_build_tnl_gtpu(struct dr_ste_ctx *ste_ctx,
+			   struct dr_ste_build *sb,
 			   struct dr_match_param *mask,
 			   bool inner, bool rx);
-void dr_ste_build_general_purpose(struct dr_ste_build *sb,
+void dr_ste_build_general_purpose(struct dr_ste_ctx *ste_ctx,
+				  struct dr_ste_build *sb,
 				  struct dr_match_param *mask,
 				  bool inner, bool rx);
-void dr_ste_build_register_0(struct dr_ste_build *sb,
+void dr_ste_build_register_0(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_register_1(struct dr_ste_build *sb,
+void dr_ste_build_register_1(struct dr_ste_ctx *ste_ctx,
+			     struct dr_ste_build *sb,
 			     struct dr_match_param *mask,
 			     bool inner, bool rx);
-void dr_ste_build_src_gvmi_qpn(struct dr_ste_build *sb,
+void dr_ste_build_src_gvmi_qpn(struct dr_ste_ctx *ste_ctx,
+			       struct dr_ste_build *sb,
 			       struct dr_match_param *mask,
 			       struct dr_devx_caps *caps,
 			       bool inner, bool rx);
+void dr_ste_build_flex_parser_0(struct dr_ste_ctx *ste_ctx,
+				struct dr_ste_build *sb,
+				struct dr_match_param *mask,
+				bool inner, bool rx);
+void dr_ste_build_flex_parser_1(struct dr_ste_ctx *ste_ctx,
+				struct dr_ste_build *sb,
+				struct dr_match_param *mask,
+				bool inner, bool rx);
 void dr_ste_build_empty_always_hit(struct dr_ste_build *sb, bool rx);
 
 /* Actions utils */
@@ -544,21 +624,34 @@ struct dr_match_misc3 {
 	uint32_t outer_vxlan_gpe_next_protocol:8;
 	uint32_t icmpv4_header_data;
 	uint32_t icmpv6_header_data;
-	uint32_t icmpv6_code:8;
-	uint32_t icmpv6_type:8;
-	uint32_t icmpv4_code:8;
-	uint32_t icmpv4_type:8;
+	uint8_t icmpv6_code;
+	uint8_t icmpv6_type;
+	uint8_t icmpv4_code;
+	uint8_t icmpv4_type;
+	uint32_t geneve_tlv_option_0_data;
 	uint32_t gtpu_teid;
 	uint32_t gtpu_msg_type:8;
 	uint32_t gtpu_flags:3;
 };
 
+struct dr_match_misc4 {
+	uint32_t prog_sample_field_value_0;
+	uint32_t prog_sample_field_id_0;
+	uint32_t prog_sample_field_value_1;
+	uint32_t prog_sample_field_id_1;
+	uint32_t prog_sample_field_value_2;
+	uint32_t prog_sample_field_id_2;
+	uint32_t prog_sample_field_value_3;
+	uint32_t prog_sample_field_id_3;
+};
+
 struct dr_match_param {
 	struct dr_match_spec	outer;
 	struct dr_match_misc	misc;
 	struct dr_match_spec	inner;
 	struct dr_match_misc2	misc2;
 	struct dr_match_misc3	misc3;
+	struct dr_match_misc4	misc4;
 };
 
 #define DR_MASK_IS_ICMPV4_SET(_misc3) ((_misc3)->icmpv4_type || \
@@ -571,6 +664,7 @@ struct dr_esw_caps {
 	uint64_t uplink_icm_address_rx;
 	uint64_t uplink_icm_address_tx;
 	bool sw_owner;
+	bool sw_owner_v2;
 };
 
 struct dr_devx_vport_cap {
@@ -580,6 +674,8 @@ struct dr_devx_vport_cap {
 };
 
 struct dr_devx_roce_cap {
+	bool roce_en;
+	bool fl_rc_qp_when_roce_disabled;
 	bool fl_rc_qp_when_roce_enabled;
 };
 
@@ -591,26 +687,92 @@ struct dr_devx_caps {
 	uint64_t			esw_rx_drop_address;
 	uint64_t			esw_tx_drop_address;
 	uint32_t			log_icm_size;
+	uint8_t				log_modify_hdr_icm_size;
 	uint64_t			hdr_modify_icm_addr;
 	uint32_t			flex_protocols;
 	uint8_t				flex_parser_id_icmp_dw0;
 	uint8_t				flex_parser_id_icmp_dw1;
 	uint8_t				flex_parser_id_icmpv6_dw0;
 	uint8_t				flex_parser_id_icmpv6_dw1;
+	uint8_t				flex_parser_id_geneve_opt_0;
+	uint8_t				flex_parser_id_mpls_over_gre;
+	uint8_t				flex_parser_id_mpls_over_udp;
 	uint8_t				max_ft_level;
+	uint8_t				sw_format_ver;
+	bool				isolate_vl_tc;
 	bool				eswitch_manager;
 	bool				rx_sw_owner;
 	bool				tx_sw_owner;
 	bool				fdb_sw_owner;
+	bool				rx_sw_owner_v2;
+	bool				tx_sw_owner_v2;
+	bool				fdb_sw_owner_v2;
 	uint32_t			num_vports;
 	struct dr_devx_vport_cap	*vports_caps;
 	struct dr_devx_roce_cap		roce_caps;
+	bool				prio_tag_required;
+};
+
+struct dr_devx_flow_table_attr {
+	uint8_t		type;
+	uint8_t		level;
+	bool		sw_owner;
+	bool		term_tbl;
+	bool		reformat_en;
+	uint64_t	icm_addr_rx;
+	uint64_t	icm_addr_tx;
+};
+
+struct dr_devx_flow_group_attr {
+	uint32_t	table_id;
+	uint32_t	table_type;
+};
+
+struct dr_devx_flow_dest_info {
+	enum dr_devx_flow_dest_type type;
+	union {
+		uint32_t vport_num;
+		uint32_t tir_num;
+		uint32_t counter_id;
+	};
+	bool has_reformat;
+	uint32_t reformat_id;
+};
+
+struct dr_devx_flow_fte_attr {
+	uint32_t			table_id;
+	uint32_t			table_type;
+	uint32_t			group_id;
+	uint32_t			flow_tag;
+	uint32_t			action;
+	uint32_t			dest_size;
+	struct dr_devx_flow_dest_info	*dest_arr;
+	bool				extended_dest;
+};
+
+struct dr_devx_tbl {
+	uint8_t			type;
+	uint8_t			level;
+	struct mlx5dv_devx_obj	*ft_dvo;
+	struct mlx5dv_devx_obj	*fg_dvo;
+	struct mlx5dv_devx_obj	*fte_dvo;
+};
+
+struct dr_devx_flow_sampler_attr {
+	uint8_t		table_type;
+	uint8_t		level;
+	uint8_t		ignore_flow_level;
+	uint32_t	sample_ratio;
+	uint32_t	default_next_table_id;
+	uint32_t	sample_table_id;
 };
 
 struct dr_domain_rx_tx {
 	uint64_t		drop_icm_addr;
 	uint64_t		default_icm_addr;
 	enum dr_ste_entry_type	ste_type;
+	/* protect rx/tx domain */
+	pthread_mutex_t		mutex;
 };
 
 struct dr_domain_info {
@@ -631,11 +793,11 @@ enum dr_domain_flags {
 
 struct mlx5dv_dr_domain {
 	struct ibv_context		*ctx;
+	struct dr_ste_ctx		*ste_ctx;
 	struct ibv_pd			*pd;
 	struct mlx5dv_devx_uar		*uar;
 	enum mlx5dv_dr_domain_type	type;
 	atomic_int			refcount;
-	pthread_mutex_t			mutex;
 	struct dr_icm_pool		*ste_icm_pool;
 	struct dr_icm_pool		*action_icm_pool;
 	struct dr_send_ring		*send_ring;
@@ -644,6 +806,28 @@ struct mlx5dv_dr_domain {
 	uint32_t			flags;
 };
 
+static inline void dr_domain_nic_lock(struct dr_domain_rx_tx *nic_dmn)
+{
+	pthread_mutex_lock(&nic_dmn->mutex);
+}
+
+static inline void dr_domain_nic_unlock(struct dr_domain_rx_tx *nic_dmn)
+{
+	pthread_mutex_unlock(&nic_dmn->mutex);
+}
+
+static inline void dr_domain_lock(struct mlx5dv_dr_domain *dmn)
+{
+	dr_domain_nic_lock(&dmn->info.rx);
+	dr_domain_nic_lock(&dmn->info.tx);
+}
+
+static inline void dr_domain_unlock(struct mlx5dv_dr_domain *dmn)
+{
+	dr_domain_nic_unlock(&dmn->info.tx);
+	dr_domain_nic_unlock(&dmn->info.rx);
+}
+
 struct dr_table_rx_tx {
 	struct dr_ste_htbl		*s_anchor;
 	struct dr_domain_rx_tx		*nic_dmn;
@@ -683,12 +867,33 @@ struct mlx5dv_dr_matcher {
 	struct list_head		rule_list;
 };
 
-struct dr_rule_member {
-	struct dr_ste		*ste;
-	/* attached to dr_rule via this */
-	struct list_node	list;
-	/* attached to dr_ste via this */
-	struct list_node	use_ste_list;
+struct dr_ste_action_modify_field {
+	uint16_t hw_field;
+	uint8_t start;
+	uint8_t end;
+	uint8_t l3_type;
+	uint8_t l4_type;
+};
+
+struct dr_devx_tbl_with_refs {
+	uint16_t		ref_actions_num;
+	struct mlx5dv_dr_action	**ref_actions;
+	struct dr_devx_tbl	*devx_tbl;
+};
+
+struct dr_flow_sampler {
+	struct mlx5dv_devx_obj	*devx_obj;
+	uint64_t		rx_icm_addr;
+	uint64_t		tx_icm_addr;
+	struct mlx5dv_dr_table	*next_ft;
+};
+
+struct dr_flow_sampler_restore_tbl {
+	struct mlx5dv_dr_table		*tbl;
+	struct mlx5dv_dr_matcher	*matcher;
+	struct mlx5dv_dr_rule		*rule;
+	struct mlx5dv_dr_action		**actions;
+	uint16_t			num_of_actions;
 };
 
 struct mlx5dv_dr_action {
@@ -728,7 +933,21 @@ struct mlx5dv_dr_action {
 			uint64_t		rx_icm_addr;
 			uint64_t		tx_icm_addr;
 		} meter;
+		struct {
+			struct mlx5dv_dr_domain			*dmn;
+			struct dr_devx_tbl_with_refs		*term_tbl;
+			struct dr_flow_sampler			*sampler_default;
+			struct dr_flow_sampler_restore_tbl	*restore_tbl;
+			struct dr_flow_sampler			*sampler_restore;
+		} sampler;
 		struct mlx5dv_dr_table	*dest_tbl;
+		struct {
+			struct mlx5dv_dr_domain		*dmn;
+			struct list_head		actions_list;
+			struct dr_devx_tbl		*devx_tbl;
+			uint64_t			rx_icm_addr;
+			uint64_t			tx_icm_addr;
+		} dest_array;
 		struct {
 			struct mlx5dv_devx_obj	*devx_obj;
 			uint32_t		offset;
@@ -738,6 +957,9 @@ struct mlx5dv_dr_action {
 			struct dr_devx_vport_cap	*caps;
 			uint32_t			num;
 		} vport;
+		struct {
+			uint32_t	vlan_hdr;
+		} push_vlan;
 		struct {
 			bool    is_qp;
 			union {
@@ -745,6 +967,7 @@ struct mlx5dv_dr_action {
 				struct ibv_qp           *qp;
 			};
 		} dest_qp;
+		struct dr_action_aso	aso;
 		struct mlx5dv_devx_obj	*devx_obj;
 		uint32_t		flow_tag;
 	};
@@ -768,10 +991,9 @@ struct dr_htbl_connect_info {
 	};
 };
 
-
 struct dr_rule_rx_tx {
-	struct list_head		rule_members_list;
 	struct dr_matcher_rx_tx		*nic_matcher;
+	struct dr_ste			*last_rule_ste;
 };
 
 struct mlx5dv_dr_rule {
@@ -787,7 +1009,13 @@ struct mlx5dv_dr_rule {
 	struct list_node	rule_list;
 };
 
-void dr_rule_update_rule_member(struct dr_ste *new_ste, struct dr_ste *ste);
+void dr_rule_set_last_member(struct dr_rule_rx_tx *nic_rule,
+			     struct dr_ste *ste,
+			     bool force);
+
+void dr_rule_get_reverse_rule_members(struct dr_ste **ste_arr,
+				      struct dr_ste *curr_ste,
+				      int *num_of_stes);
 
 struct dr_icm_chunk {
 	struct dr_icm_buddy_mem *buddy_mem;
@@ -860,11 +1088,22 @@ int dr_devx_query_gvmi(struct ibv_context *ctx,
 int dr_devx_query_esw_caps(struct ibv_context *ctx,
 			   struct dr_esw_caps *caps);
 int dr_devx_sync_steering(struct ibv_context *ctx);
-struct mlx5dv_devx_obj *dr_devx_create_flow_table(struct ibv_context *ctx,
-						  uint32_t table_type,
-						  uint64_t icm_addr_rx,
-						  uint64_t icm_addr_tx,
-						  u8 level);
+struct mlx5dv_devx_obj *
+dr_devx_create_flow_table(struct ibv_context *ctx,
+			  struct dr_devx_flow_table_attr *table_attr);
+int dr_devx_query_flow_table(struct mlx5dv_devx_obj *obj,  uint32_t type,
+			     uint64_t *rx_icm_addr, uint64_t *tx_icm_addr);
+struct dr_devx_tbl *
+dr_devx_create_always_hit_ft(struct ibv_context *ctx,
+			     struct dr_devx_flow_table_attr *ft_attr,
+			     struct dr_devx_flow_group_attr *fg_attr,
+			     struct dr_devx_flow_fte_attr *fte_attr);
+void dr_devx_destroy_always_hit_ft(struct dr_devx_tbl *devx_tbl);
+struct mlx5dv_devx_obj *
+dr_devx_create_flow_sampler(struct ibv_context *ctx,
+			    struct dr_devx_flow_sampler_attr *sampler_attr);
+int dr_devx_query_flow_sampler(struct mlx5dv_devx_obj *obj,
+			       uint64_t *rx_icm_addr, uint64_t *tx_icm_addr);
 struct mlx5dv_devx_obj *dr_devx_create_reformat_ctx(struct ibv_context *ctx,
 						    enum reformat_type rt,
 						    size_t reformat_size,
@@ -896,6 +1135,7 @@ struct dr_devx_qp_create_attr {
 	uint32_t	sq_wqe_cnt;
 	uint32_t	rq_wqe_cnt;
 	uint32_t	rq_wqe_shift;
+	bool		isolate_vl_tc;
 };
 
 struct mlx5dv_devx_obj *dr_devx_create_qp(struct ibv_context *ctx,
@@ -945,16 +1185,20 @@ static inline bool dr_is_root_table(struct mlx5dv_dr_table *tbl)
 struct dr_icm_pool *dr_icm_pool_create(struct mlx5dv_dr_domain *dmn,
 				       enum dr_icm_type icm_type);
 void dr_icm_pool_destroy(struct dr_icm_pool *pool);
+int dr_icm_pool_sync_pool(struct dr_icm_pool *pool);
 
 struct dr_icm_chunk *dr_icm_alloc_chunk(struct dr_icm_pool *pool,
 					enum dr_icm_chunk_size chunk_size);
 void dr_icm_free_chunk(struct dr_icm_chunk *chunk);
+void dr_ste_prepare_for_postsend(struct dr_ste_ctx *ste_ctx,
+				 uint8_t *hw_ste_p, uint32_t ste_size);
 int dr_ste_htbl_init_and_postsend(struct mlx5dv_dr_domain *dmn,
 				  struct dr_domain_rx_tx *nic_dmn,
 				  struct dr_ste_htbl *htbl,
 				  struct dr_htbl_connect_info *connect_info,
 				  bool update_hw_ste);
-void dr_ste_set_formated_ste(uint16_t gvmi,
+void dr_ste_set_formated_ste(struct dr_ste_ctx *ste_ctx,
+			     uint16_t gvmi,
 			     struct dr_domain_rx_tx *nic_dmn,
 			     struct dr_ste_htbl *htbl,
 			     uint8_t *formated_ste,
@@ -991,6 +1235,7 @@ struct dr_qp {
 	struct mlx5dv_devx_uar		*uar;
 	struct mlx5dv_devx_umem		*buf_umem;
 	struct mlx5dv_devx_umem		*db_umem;
+	uint8_t nc_uar : 1;
 };
 
 struct dr_cq {
@@ -1019,6 +1264,8 @@ struct dr_send_ring {
 	uint32_t		max_post_send_size;
 	/* manage the send queue */
 	uint32_t		tx_head;
+	/* protect QP/CQ operations */
+	pthread_mutex_t         mutex;
 	void			*buf;
 	uint32_t		buf_size;
 	struct ibv_wc		wc[MAX_SEND_CQE];
@@ -1029,6 +1276,7 @@ struct dr_send_ring {
 int dr_send_ring_alloc(struct mlx5dv_dr_domain *dmn);
 void dr_send_ring_free(struct dr_send_ring *send_ring);
 int dr_send_ring_force_drain(struct mlx5dv_dr_domain *dmn);
+bool dr_send_allow_fl(struct dr_devx_caps *caps);
 int dr_send_postsend_ste(struct mlx5dv_dr_domain *dmn, struct dr_ste *ste,
 			 uint8_t *data, uint16_t size, uint16_t offset);
 int dr_send_postsend_htbl(struct mlx5dv_dr_domain *dmn, struct dr_ste_htbl *htbl,
