@@ -1,11 +1,14 @@
 # SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
 # Copyright (c) 2018 Mellanox Technologies, Inc. All rights reserved. See COPYING file
+# Copyright 2020 Amazon.com, Inc. or its affiliates. All rights reserved.
 """
 Test module for pyverbs' device module.
 """
 import unittest
 import resource
 import random
+import errno
+import os
 
 from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
 from tests.base import PyverbsAPITestCase
@@ -21,12 +24,6 @@ class DeviceTest(PyverbsAPITestCase):
     Test various functionalities of the Device class.
     """
 
-    def setUp(self):
-        pass
-
-    def tearDown(self):
-        pass
-
     def get_device_list(self):
         lst = d.get_device_list()
         if len(lst) == 0:
@@ -63,6 +60,15 @@ class DeviceTest(PyverbsAPITestCase):
                 attr = ctx.query_device()
                 self.verify_device_attr(attr, dev)
 
+    def test_query_pkey(self):
+        """
+        Test ibv_query_pkey()
+        """
+        for dev in self.get_device_list():
+            with d.Context(name=dev.name.decode()) as ctx:
+                if dev.node_type == e.IBV_NODE_CA:
+                    ctx.query_pkey(port_num=1, index=0)
+
     def test_query_gid(self):
         """
         Test ibv_query_gid()
@@ -71,6 +77,91 @@ class DeviceTest(PyverbsAPITestCase):
             with d.Context(name=dev.name.decode()) as ctx:
                 ctx.query_gid(port_num=1, index=0)
 
+    def test_query_gid_table(self):
+        """
+        Test ibv_query_gid_table()
+        """
+        devs = self.get_device_list()
+        with d.Context(name=devs[0].name.decode()) as ctx:
+            device_attr = ctx.query_device()
+            max_entries = 0
+            for port_num in range(1, device_attr.phys_port_cnt + 1):
+                port_attr = ctx.query_port(port_num)
+                max_entries += port_attr.gid_tbl_len
+            try:
+                ctx.query_gid_table(max_entries)
+            except PyverbsRDMAError as ex:
+                if ex.error_code in [-errno.EOPNOTSUPP, -errno.EPROTONOSUPPORT]:
+                    raise unittest.SkipTest('ibv_query_gid_table is not'\
+                                            ' supported on this device')
+                raise ex
+
+    def test_query_gid_table_bad_flow(self):
+        """
+        Test ibv_query_gid_table() with too small a buffer
+        """
+        ctx, _, _ = self.devices[0]
+        try:
+            ctx.query_gid_table(0)
+        except PyverbsRDMAError as ex:
+            if ex.error_code in [-errno.EOPNOTSUPP, -errno.EPROTONOSUPPORT]:
+                raise unittest.SkipTest('ibv_query_gid_table is not'
+                                        ' supported on this device')
+            self.assertEqual(ex.error_code, -errno.EINVAL,
+                             f'Got -{os.strerror(-ex.error_code)} but '
+                             f'Expected -{os.strerror(errno.EINVAL)} ')
+        else:
+            raise PyverbsRDMAError('Successfully queried '
+                                   'gid_table with an insufficient buffer')
+
+    def test_query_gid_ex(self):
+        """
+        Test ibv_query_gid_ex()
+        """
+        devs = self.get_device_list()
+        with d.Context(name=devs[0].name.decode()) as ctx:
+            try:
+                ctx.query_gid_ex(port_num=1, gid_index=0)
+            except PyverbsRDMAError as ex:
+                if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
+                    raise unittest.SkipTest('ibv_query_gid_ex is not'\
+                                            ' supported on this device')
+                raise ex
+
+    def test_query_gid_ex_bad_flow(self):
+        """
+        Test ibv_query_gid_ex() with an empty index
+        """
+        ctx, device_attr, _ = self.devices[0]
+        try:
+            port_attr = ctx.query_port(1)
+            max_entries = 0
+            for port_num in range(1, device_attr.phys_port_cnt + 1):
+                attr = ctx.query_port(port_num)
+                max_entries += attr.gid_tbl_len
+            gid_indices = {gid_entry.gid_index for gid_entry in
+                           ctx.query_gid_table(max_entries) if gid_entry.port_num == 1}
+
+            possible_indices = set(range(port_attr.gid_tbl_len)) if port_attr.gid_tbl_len > 1 else set()
+            try:
+                no_gid_index = possible_indices.difference(gid_indices).pop()
+            except KeyError:
+                # all indices are populated by GIDs
+                raise unittest.SkipTest('All gid indices populated,'
+                                        ' cannot check bad flow')
+
+            ctx.query_gid_ex(port_num=1, gid_index=no_gid_index)
+        except PyverbsRDMAError as ex:
+            if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
+                raise unittest.SkipTest('ibv_query_gid_ex is not'
+                                        ' supported on this device')
+            self.assertEqual(ex.error_code, errno.ENODATA,
+                             f'Got {os.strerror(ex.error_code)} but '
+                             f'Expected {os.strerror(errno.ENODATA)}')
+        else:
+            raise PyverbsRDMAError('Successfully queried '
+                                   f'non-existent gid index {no_gid_index}')
+
     @staticmethod
     def verify_device_attr(attr, device):
         """
@@ -86,16 +177,16 @@ class DeviceTest(PyverbsAPITestCase):
         assert attr.max_mr_size > PAGE_SIZE
         assert attr.page_size_cap >= PAGE_SIZE
         assert attr.vendor_id != 0
-        assert attr.vendor_part_id != 0
         assert attr.max_qp > 0
         assert attr.max_qp_wr > 0
         assert attr.max_sge > 0
-        assert attr.max_sge_rd > 0
+        assert attr.max_sge_rd >= 0
         assert attr.max_cq > 0
         assert attr.max_cqe > 0
         assert attr.max_mr > 0
         assert attr.max_pd > 0
-        assert attr.max_pkeys > 0
+        if device.node_type == e.IBV_NODE_CA:
+            assert attr.max_pkeys > 0
 
     def test_query_device_ex(self):
         """
